/*--------------------------------------------------------------------------
Program     : utl/qry/qry-mgr.sp

Description : Query management code

Copyright(c): Taylor Steel Inc. 2004 Taylor Steel Inc.
Author      : Tim Kuehn
Created     : 8/24/2004 TAYLOR STEEL INC
Notes       :
       
--------Revision History---------------------------------------------------

	Last change: TDK 2006-08-08 10:15:26
--------------------------------------------------------------------------*/

    /* set the following to FALSE                                           */
    /* if the procedure manager is not in use                               */

&SCOPED-DEFINE is-use-procedure-manager     TRUE

/****************************************************************************/
/* External Files */

&IF DEFINED(super-header) = 0 &THEN             /* Things that only belong in a SP          */

&IF ({&is-use-procedure-manager}) &THEN         /* This is being used with the procedure    */
                                                /* manager                                  */
    /* Procedure manager configuration  */

{utl/ppmgr/spopen-public-unique-permanent.i }   /* Standard procedure manager SP header     */

&ENDIF

/****************************************************************************/
/* Super Procedure Headers              */

{utl/qry/qry-mgr.i      }           /* Definitions                          */

/****************************************************************************/
/* External Temp Table Defs             */

{utl/qry/qry-mgr.tt     }           /* Temp tables                          */

/****************************************************************************/
/* Memory Managment                     */

CREATE WIDGET-POOL.

/****************************************************************************/
/* Local Variables */

DEFINE VARIABLE cur-local-name                          AS CHARACTER NO-UNDO.       /* Names for things I need locally              */
DEFINE VARIABLE cur-name-seq                            AS INTEGER   NO-UNDO.       /* Sequence for generating unique names         */
                                                                                    /* for internal use only                        */
&ENDIF

/****************************************************************************/
FUNCTION get-qrymgr-quoted-string
/****************************************************************************
PURPOSE:

    Takes a string and returns it ready for use in a query string specification.

    This turns

            this is a "String"


    into

            "this is a ""string"""

    This value can then be used to  construct the string components of a query
    search specification like so:

        cur-query-string = cur-query-string +
                           get-qrymgr-quoted-string("this is a ""string"".").

PARAMETERS:

    cur-string      String to prepare for use in a query condition

*****************************************************************************/
    RETURNS CHARACTER

    (INPUT cur-string AS CHARACTER)                         {&super-header}:

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Variables */

DEFINE VARIABLE cur-char                                AS CHARACTER NO-UNDO.

DEFINE VARIABLE cur-quote-string-list                   AS CHARACTER NO-UNDO
                                                            INITIAL "~~""~\.".

DEFINE VARIABLE cur-quote-char                          AS CHARACTER NO-UNDO.

DEFINE VARIABLE cur-int                                 AS INTEGER   NO-UNDO.

/****************************************************************************/

ASSIGN
    cur-char = cur-string
    .

/****************************************************************************/

    /* If this isn't a missing value    */

IF cur-char <> ? THEN
    DO:

        /* Quote things out the way they should             */

    DO cur-int = 1 TO LENGTH(cur-quote-string-list):

        ASSIGN
            cur-quote-char  = SUBSTRING(cur-quote-string-list, cur-int, 1)              /* Get the quoteable char   */
            cur-char        = REPLACE(cur-char, cur-quote-char, "~~" + cur-quote-char)  /* Then quote it            */
            .

    END.

        /* Surround the string with quotes                  */

    ASSIGN
        cur-char =  CHR(34) + cur-char + CHR(34)
        .

    END.


ELSE    /* This is a missing value - turn it into a         */
        /* missing value usable by the query code           */

    ASSIGN
        cur-char = "?"
        .

RETURN(cur-char).
END FUNCTION.
&ENDIF

/****************************************************************************/
PROCEDURE create-qrymgr-query                               {&super-header}:
/****************************************************************************
PURPOSE:

    Create a new dynamic query with an associated set of either new buffers
    or previously existing buffers.

PARAMETERS:

    cur-table-name-handle-list  Comma-seperated list of table names and/or
                                buffer-handles

                                The format is
                                
                                    STRING(BUFFER cumast:HANDLE) + ",cushipto"

                                The position of the buffer specifications are
                                the same as how they're added to a query.

    cur-flag-list               Optional flags to control behavior.

        Flag                    Description
        ---------------         ---------------------------------
        "reposition"            This is a scrolling query, so forward-only = false  (9.1D and following versions only)

        "browse"                This is a scrolling query, so forward-only = false  (9.1D and following versions only)
                                If "browse" is specified, and tt-query.browse-handle
                                is a valid handle, then this code will also automatically
                                refresh the browse object when the query's opened or reopened

        "force-buffer-create"   Always create a new buffer for all passed-in table references
                                regardless of whether this is a valid buffer-handle or not.

                                This is used for situations where a valid buffer handle's
                                been passed in, but a new buffer's required so the original
                                buffer the handle points to isn't disturbed

    cur-query-handle            OUTPUT parameter with the handle to the
                                newly created query

*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-table-name-handle-list       AS CHARACTER NO-UNDO.
DEFINE INPUT PARAMETER cur-flag-list                    AS CHARACTER NO-UNDO.

DEFINE OUTPUT PARAMETER cur-query-handle                AS HANDLE    NO-UNDO.

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Variables          */

DEFINE VARIABLE cur-entry-num                           AS INTEGER   NO-UNDO.
DEFINE VARIABLE cur-entry-value                         AS CHARACTER NO-UNDO.
DEFINE VARIABLE cur-buf-handle                          AS HANDLE    NO-UNDO.

DEFINE VARIABLE cur-tmp-handle                          AS HANDLE    NO-UNDO.

DEFINE VARIABLE cur-num-entries                         AS INTEGER   NO-UNDO.

DEFINE VARIABLE is-force-buffer-create                  AS LOGICAL   NO-UNDO.

/****************************************************************************/

DO TRANSACTION: /* Transaction */

/****************************************************************************/
/* Create the query and     */
/* save the search spec     */

CREATE QUERY cur-query-handle.

CREATE tt-query.

ASSIGN
    tt-query.qry-handle             = cur-query-handle

    tt-query.browse-handle          = ?
    tt-query.is-browse-control      = LOOKUP("browse",              cur-flag-list) > 0

    tt-query.is-forward-only        = LOOKUP("reposition",          cur-flag-list) = 0 AND
                                      LOOKUP("browse",              cur-flag-list) = 0

    is-force-buffer-create          = LOOKUP("force-buffer-create", cur-flag-list) > 0
    .

&IF KEYWORD-ALL( "FORWARD-ONLY" ) NE ? &THEN
ASSIGN
    cur-query-handle:FORWARD-ONLY   = tt-query.is-forward-only
    .
&ENDIF

/****************************************************************************/
/* Extract or create the buffer handles */

ASSIGN
    cur-num-entries = NUM-ENTRIES(cur-table-name-handle-list)
    .

DO cur-entry-num = 1 TO cur-num-entries:

    ASSIGN
        cur-entry-value = ENTRY(cur-entry-num, cur-table-name-handle-list)
        .

        /* Try to turn this into a buffer handle    */

    IF NOT is-force-buffer-create THEN 

        ASSIGN
            cur-tmp-handle  = WIDGET-HANDLE(cur-entry-value)
            NO-ERROR
            .

    ELSE

        ASSIGN
            cur-tmp-handle = ?
            .

        /* Get the buffer handle reference          */
        
    IF VALID-HANDLE(cur-tmp-handle) THEN                                /* A usable handle was passed in        */
        DO:

        RUN create-tt-query-buffer-record(  BUFFER tt-query-buffer,
                                            tt-query.qry-handle,
                                            cur-entry-num,
                                            WIDGET-HANDLE(cur-entry-value),
                                            NO).

        END.

    ELSE                                                                /* otherwise                            */

        DO:

        CREATE BUFFER cur-buf-handle                                    /* assume this is a a [db.]table name   */
            FOR TABLE cur-entry-value.

        RUN create-tt-query-buffer-record(  BUFFER tt-query-buffer,
                                            tt-query.qry-handle,
                                            cur-entry-num,
                                            cur-buf-handle,
                                            YES).


        END.

        /* Add the buffers to the query */

    cur-query-handle:ADD-BUFFER(tt-query-buffer.buffer-handle).
        
END.

END.            /* Transaction */

&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE import-qrymgr-query                               {&super-header}:
/****************************************************************************
PURPOSE:

    Import an externally defined query into the manager

PARAMETERS:

    cur-query-handle    Handle to the static query

    cur-flag-list       Optional flags to control behavior.

                        Flag            Description
                        --------------- ---------------------------------
                        "browse"        If "browse" is specified, and tt-query.browse-handle
                                        is a valid handle, then this code will also automatically
                                        refresh the browse object when the query's opened or reopened


*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE    NO-UNDO.
DEFINE INPUT PARAMETER cur-flag-list                    AS CHARACTER NO-UNDO.

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Variables */

DEFINE VARIABLE cur-cnt                                 AS INTEGER   NO-UNDO.
DEFINE VARIABLE cur-buffer-handle                       AS HANDLE    NO-UNDO.

/****************************************************************************/

DO TRANSACTION: /* Transaction */

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

FIND tt-query                                                   /* Check if the query's registered  */
    WHERE tt-query.qry-handle = cur-query-handle
    EXCLUSIVE-LOCK
    NO-ERROR.

IF AVAILABLE tt-query THEN                                      /* If it is, then we're done        */
    LEAVE.

CREATE tt-query.

ASSIGN
    tt-query.qry-handle         = cur-query-handle
    tt-query.qry-name           = cur-query-handle:NAME

    tt-query.is-created-local   = NO

    tt-query.is-browse-control  = LOOKUP("browse", cur-flag-list) > 0
    .

&IF KEYWORD-ALL( "FORWARD-ONLY" ) NE ? &THEN
    tt-query.is-forward-only    = cur-query-handle:FORWARD-ONLY
&ENDIF
    .

DO cur-cnt = 1 TO cur-query-handle:NUM-BUFFERS:

    RUN create-tt-query-buffer-record(  BUFFER tt-query-buffer,
                                        cur-query-handle,
                                        cur-cnt,
                                        cur-query-handle:GET-BUFFER-HANDLE(cur-cnt),
                                        NO).
END.

END.            /* Transaction */

&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE set-qrymgr-query-browse-handle                    {&super-header}:
/****************************************************************************
PURPOSE:

    Given a query handle and browse handle, associate the browse with
    the query in the tt-query table.

PARAMETERS:

    cur-query-handle    Handle to the associated query

    cur-browse-handle   Handle to the associated browse

*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE    NO-UNDO.
DEFINE INPUT PARAMETER cur-browse-handle                AS HANDLE    NO-UNDO.

/****************************************************************************/

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/

DO TRANSACTION: /* Transaction */

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

FIND tt-query
    WHERE tt-query.qry-handle = cur-query-handle
    EXCLUSIVE-LOCK
    NO-ERROR.

IF AVAILABLE tt-query THEN
    DO:

    ASSIGN
        tt-query.browse-handle                  = cur-browse-handle

        tt-query.parent-browse-handle           = ?
        tt-query.parent-query-handle            = ?

            /* Default values for these     */

        tt-query.is-select-focused-row-pending  = YES
        tt-query.is-select-row-1-pending        = NO
        .

    FIND CURRENT tt-query
        NO-LOCK.

    END.

RUN control-browse(BUFFER tt-query, "enable").

END.            /* Transaction */

&ENDIF
END PROCEDURE.


/****************************************************************************/
PROCEDURE import-qrymgr-browse                              {&super-header}:
/****************************************************************************
PURPOSE:

    Import a browse and it's associated query, and set other values
    to sane defaults

PARAMETERS:

    cur-browse-handle       Handle to the browse object to import

*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-browse-handle                AS HANDLE    NO-UNDO.

/****************************************************************************/

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Variables */

DEFINE VARIABLE cur-query-handle                        AS HANDLE    NO-UNDO.

/****************************************************************************/

IF NOT VALID-HANDLE(cur-browse-handle) THEN
    LEAVE.

    /* Bring the query in for this browse       */

RUN import-qrymgr-query(cur-browse-handle:QUERY, "browse").

    /* Then associate the browse with the query */

RUN set-qrymgr-query-browse-handle(cur-browse-handle:QUERY, cur-browse-handle).


&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE refresh-qrymgr-browse                             {&super-header}:
/****************************************************************************
PURPOSE:

    Refresh the browse associated with a query

PARAMETERS:

    cur-query-handle        Handle to the query to refresh the browse for

*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE    NO-UNDO.

/****************************************************************************/

&IF DEFINED(super-header) = 0 &THEN

/****************************************************************************/
/* Local Buffers */

DEFINE BUFFER tt-query FOR tt-query.

/****************************************************************************/

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

FIND tt-query                                           /* Find the query                   */
    WHERE tt-query.qry-handle = cur-query-handle
    NO-LOCK
    NO-ERROR.

RUN control-browse(BUFFER tt-query, "refresh").         /* Then refresh it                  */

&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE select-focused-row-qrymgr-browse                  {&super-header}:
/****************************************************************************
PURPOSE:


PARAMETERS:

*****************************************************************************/

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE    NO-UNDO.

/****************************************************************************/

&IF DEFINED(super-header) = 0 &THEN

/****************************************************************************/
/* Local Buffers */

DEFINE BUFFER tt-query FOR tt-query.

/****************************************************************************/

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

FIND tt-query                                           /* Find the query   */
    WHERE tt-query.qry-handle = cur-query-handle
    NO-LOCK
    NO-ERROR.

RUN control-browse(BUFFER tt-query, "select-focused-row").

&ENDIF
END PROCEDURE.

/****************************************************************************/
FUNCTION apply-event-to-qrymgr-browse
/****************************************************************************
PURPOSE:

    Given a

        CURSOR-UP,  CURSOR-DOWN,
        PAGE-UP,    PAGE-DOWN

    event applied to a browse, make sure when we're
    done the selected row's still has focus.

    Return a logical to indicate where to do a RETURN NO-APPLY or not.

    YES - apply the event
    NO  - RETURN NO-APPLY.

PARAMETERS:

    cur-browse-handle           Handle to the browse that's moving

    cur-last-event-function     LAST-EVENT:FUNCTION that actually happened

*****************************************************************************/
     RETURNS LOGICAL

     (INPUT cur-browse-handle       AS HANDLE,
      INPUT cur-last-event-function AS CHARACTER)           {&super-header}:

&IF DEFINED(super-header) = 0 &THEN

/****************************************************************************/
/* Local Variables */

DEFINE VARIABLE is-logical                              AS LOGICAL   NO-UNDO.

/*****************************************************************************/

ASSIGN
    is-logical = YES
    .

DO WHILE TRUE:

        /* Was this aritificially generated?    */
        /* Don't do anything if it was          */

    IF NOT (cur-last-event-function BEGINS "cursor" OR
            cur-last-event-function BEGINS "page")      THEN
        LEAVE.

        /* Apply the last event to this         */

    APPLY cur-last-event-function
        TO cur-browse-handle.

    IF cur-last-event-function BEGINS "page" THEN
        RUN select-focused-row-qrymgr-browse(cur-browse-handle).

    ASSIGN
        is-logical = NO
        .
        
    LEAVE.

END.

RETURN(is-logical).

END FUNCTION.
&ENDIF

/****************************************************************************/
PROCEDURE set-qrymgr-query-buffer-join-spec                 {&super-header}:
/****************************************************************************
PURPOSE:

    Set the join spec for a given query buffer. This is distinct from the
    search spec in that it's intended to only join the current join level
    buffer to a prior join level buffers. For the first level buffer, it
    is only used to specify if this is a "FIRST", "EACH", or "LAST" record
    set is required.

PARAMETERS:

    cur-query-handle        handle to the query

    cur-buffer-num          buffer position in the query

    cur-join-record-set     This is the record set to join, with an option
                            modifier for outer join types.

                            The first entry must be one of "EACH", "FIRST",
                            or "LAST".

                            The (optional) second part can be used to
                            specify if this is an outer join or left outer-join.

                            The format for outer-join type queries would be

                                "EACH,OUTER-JOIN"

    cur-join-spec           join spec to the prior level of the query in

                                &1.field = &2.field

                            format

                            Buffer "1" should only be one of

                                FOR EACH &1
                                FOR FIRST &1
                                FOR LAST &1

*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE    NO-UNDO.
DEFINE INPUT PARAMETER cur-buffer-num                   AS INTEGER   NO-UNDO.

DEFINE INPUT PARAMETER cur-join-record-set              AS CHARACTER NO-UNDO.
DEFINE INPUT PARAMETER cur-join-spec                    AS CHARACTER NO-UNDO.

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

DO TRANSACTION: /* Transaction */

FIND tt-query-buffer
    WHERE tt-query-buffer.qry-handle    = cur-query-handle  AND
          tt-query-buffer.buffer-num    = cur-buffer-num
    EXCLUSIVE-LOCK
    NO-ERROR.

IF NOT AVAILABLE tt-query-buffer THEN
    LEAVE.

ASSIGN
    tt-query-buffer.join-record-set = ENTRY(1, cur-join-record-set)

    tt-query-buffer.join-type       = IF NUM-ENTRIES(cur-join-record-set) > 1
                                        THEN ENTRY(2, cur-join-record-set) 
                                        ELSE ""

    tt-query-buffer.join-spec       = cur-join-spec
    .

RUN set-tt-query-join-search-spec(BUFFER tt-query-buffer).

FIND CURRENT tt-query-buffer
    NO-LOCK.

END.            /* Transaction */

&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE set-qrymgr-query-buffer-search-spec               {&super-header}:
/****************************************************************************
PURPOSE:

    Set the search spec for a given query buffer. This replaces any
    current specification that may be on this buffer.

PARAMETERS:

    cur-query-handle        Handle to the query

    cur-buffer-num          Buffer position in the query

    cur-search-spec         Search spec for this buffer.

*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE    NO-UNDO.
DEFINE INPUT PARAMETER cur-buffer-num                   AS INTEGER   NO-UNDO.
DEFINE INPUT PARAMETER cur-search-spec                  AS CHARACTER NO-UNDO.

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

DO TRANSACTION: /* Transaction */

FIND tt-query-buffer
    WHERE tt-query-buffer.qry-handle    = cur-query-handle  AND
          tt-query-buffer.buffer-num    = cur-buffer-num
    EXCLUSIVE-LOCK
    NO-ERROR.

IF NOT AVAILABLE tt-query-buffer THEN
    LEAVE.

ASSIGN
    tt-query-buffer.search-spec = cur-search-spec
    .

RUN set-tt-query-join-search-spec(BUFFER tt-query-buffer).

FIND CURRENT tt-query-buffer
    NO-LOCK.

END.            /* Transaction */

&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE add-qrymgr-query-buffer-compound-search-spec      {&super-header}:
/****************************************************************************
PURPOSE:

    Given a two-field component search configuration,
    construct a query spec to do range or equality matching
    for that component field setup

PARAMETERS:

    cur-query-handle            Handle to the query

    cur-buffer-num              Buffer position in the query

    cur-primary-field-name      Name of the primary field in the buffer
    cur-secondary-field-name    Name of the secondary field in the buffer

    cur-boolean-op              One of <>, =, >=, <=, >, <

    cur-primary-value-string    Value of the primary field for (range) matching
    cur-secondary-value-string  Value of the secondary field for (range) matching

    cur-flag-list               List of flags that control behavior

                                Flag                    Description
                                ----------------------  ------------------------------
                                "mandatory"             Both the primary and secondary
                                                        fields are manadatory matches
                                                        even if they're ? or ""

                                "mandatory-primary"     The primary match value is a
                                                        manadatory match  even if it's
                                                        a ? or ""

                                "mandatory-secondary"   The secondary match value is a
                                                        manadatory match  even if it's
                                                        a ? or ""


*****************************************************************************/
/* Local Parameters                     */

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE       NO-UNDO.
DEFINE INPUT PARAMETER cur-buffer-num                   AS INTEGER      NO-UNDO.

DEFINE INPUT PARAMETER cur-primary-field-name           AS CHARACTER    NO-UNDO.
DEFINE INPUT PARAMETER cur-secondary-field-name         AS CHARACTER    NO-UNDO.

DEFINE INPUT PARAMETER cur-boolean-op                   AS CHARACTER    NO-UNDO.

DEFINE INPUT PARAMETER cur-primary-value-string         AS CHARACTER    NO-UNDO.
DEFINE INPUT PARAMETER cur-secondary-value-string       AS CHARACTER    NO-UNDO.

DEFINE INPUT PARAMETER cur-flag-list                    AS CHARACTER    NO-UNDO.

/****************************************************************************/
&IF DEFINED(super-header) = 0 &THEN

/****************************************************************************/
/* Local Variables                      */

DEFINE VARIABLE is-mandatory                            AS LOGICAL      NO-UNDO.
DEFINE VARIABLE is-mandatory-primary                    AS LOGICAL      NO-UNDO.
DEFINE VARIABLE is-mandatory-secondary                  AS LOGICAL      NO-UNDO.

DEFINE VARIABLE is-primary-value-present                AS LOGICAL      NO-UNDO.
DEFINE VARIABLE is-secondary-value-present              AS LOGICAL      NO-UNDO.

DEFINE VARIABLE cur-primary-boolean-op                  AS CHARACTER    NO-UNDO.
DEFINE VARIABLE cur-secondary-boolean-op                AS CHARACTER    NO-UNDO.

DEFINE VARIABLE cur-search-spec                         AS CHARACTER    NO-UNDO.

/****************************************************************************/

ASSIGN
    is-mandatory                = LOOKUP("mandatory",           cur-flag-list) > 0

    is-mandatory-primary        = LOOKUP("mandatory-primary",   cur-flag-list) > 0  OR is-mandatory
    is-mandatory-secondary      = LOOKUP("mandatory-secondary", cur-flag-list) > 0  OR is-mandatory

    is-primary-value-present    = (cur-primary-value-string     <> ? AND cur-primary-value-string   > "")
    is-secondary-value-present  = (cur-secondary-value-string   <> ? AND cur-secondary-value-string > "")
    .

/****************************************************************************/
/* Is there anything to do here?        */

IF is-primary-value-present         AND         /* No search spec?          */
   NOT is-mandatory-primary         THEN        /* not a mandatory search   */

    LEAVE.                                      /* Then get out             */

/****************************************************************************/

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).              /* Make sure this is a query handle */

FIND tt-query-buffer
    WHERE tt-query-buffer.qry-handle    = cur-query-handle  AND
          tt-query-buffer.buffer-num    = cur-buffer-num
    NO-LOCK
    NO-ERROR.

IF NOT AVAILABLE tt-query-buffer    THEN                        /* No matching query buffer         */
    LEAVE.

/****************************************************************************/
/* Prepend the "&" buffer numbers to the table field names  */

ASSIGN
    cur-primary-field-name      = "&" + STRING(cur-buffer-num) + "." + cur-primary-field-name
    cur-secondary-field-name    = "&" + STRING(cur-buffer-num) + "." + cur-secondary-field-name
    .

/****************************************************************************/
/* Construct the primary / secondary comparison         */

CASE cur-boolean-op:

/****************************************************************************/
/* (in)equality matches                                 */

WHEN "="    OR
WHEN "<>"   THEN

    DO:

    ASSIGN
        cur-primary-value-string    =   get-qrymgr-quoted-string(cur-primary-value-string)

        cur-search-spec             =   SUBSTITUTE("&1 &2 &3",  cur-primary-field-name,
                                                                cur-boolean-op,
                                                                cur-primary-value-string)
        .

    IF is-secondary-value-present   OR
       is-mandatory-secondary       THEN

        ASSIGN
            cur-secondary-value-string  =   get-qrymgr-quoted-string(cur-secondary-value-string)

            cur-search-spec             =   cur-search-spec             +

                                            (IF cur-search-spec > ""
                                                THEN " AND "
                                                ELSE "")                +

                                            SUBSTITUTE("&1 &2 &3",  cur-secondary-field-name,
                                                                    cur-boolean-op,
                                                                    cur-secondary-value-string)
            .
    END.

/****************************************************************************/
/* All range matches    */

OTHERWISE

    DO:

        /* Range check                                      */

    ASSIGN
        cur-primary-boolean-op      =   cur-boolean-op

                /* >= or <= can only use the "<" or ">" of the range    */
                /* boolean for dual-field range comparisons             */

        cur-primary-boolean-op      =   SUBSTRING(cur-primary-boolean-op, 1, 1) WHEN    is-secondary-value-present OR
                                                                                        is-mandatory-secondary

                /* Make sure the value string is queryified                         */

        cur-primary-value-string    =   get-qrymgr-quoted-string(cur-primary-value-string)

        cur-search-spec             =   SUBSTITUTE  (   "(&1 &2 &3)",

                                                        cur-primary-field-name,
                                                        cur-primary-boolean-op,
                                                        cur-primary-value-string
                                                    )
        .

        /* Secondary field value range comparison portion   */

    IF is-secondary-value-present   OR
       is-mandatory-secondary       THEN

        DO:

        ASSIGN
            cur-secondary-value-string  =   get-qrymgr-quoted-string(cur-secondary-value-string)

            cur-search-spec             =   "("             +

                                            cur-search-spec + " OR " +

                                            SUBSTITUTE  (   "(&1 = &2 AND &4 &3 &5)",       /* For the same primary field value,                    */
                                                                                            /* do the secondary field value comparison              */

                                                            cur-primary-field-name,     cur-primary-value-string,

                                                            cur-boolean-op,                 /* >= and <= can only be done on secondary field value  */
                                                                                            /* comparisons                                          */

                                                            cur-secondary-field-name,   cur-secondary-value-string
                                                        )   +
                                            ")"
            .

        END.

    END.

END CASE.

/****************************************************************************/
/* Update the buffer search spec    */

DO TRANSACTION:   /* Transaction */

FIND CURRENT tt-query-buffer
    EXCLUSIVE-LOCK
    NO-ERROR.

RUN add-query-condition(INPUT-OUTPUT tt-query-buffer.search-spec,
                        cur-buffer-num,
                        "",
                        "APPEND",
                        cur-search-spec).

RUN set-tt-query-join-search-spec(BUFFER tt-query-buffer).

FIND CURRENT tt-query-buffer
    NO-LOCK.

END.              /* Transaction */
&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE add-qrymgr-query-buffer-search-spec-date-range    {&super-header}:
/****************************************************************************
PURPOSE:

    Add a date range to a query buffer spec

PARAMETERS:

    cur-query-handle        Handle to the query

    cur-buffer-num          Buffer position in the query

    cur-field-name          Name of the field to range match against

    cur-start-date          Starting date range.     Ignored if this is ?

    cur-end-date            End of the date range.   Ignored if this is ?

                            If cur-start-date = cur-end-date then an "=" is used.

    cur-flag-list           Optional flags to control behavior.

                            Flag            Description
                            --------------- ---------------------------------
                            "mandatory"     Don't ignore ? parameter values

*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE       NO-UNDO.
DEFINE INPUT PARAMETER cur-buffer-num                   AS INTEGER      NO-UNDO.

DEFINE INPUT PARAMETER cur-field-name                   AS CHARACTER    NO-UNDO.

DEFINE INPUT PARAMETER cur-start-date                   AS DATE         NO-UNDO.
DEFINE INPUT PARAMETER cur-end-date                     AS DATE         NO-UNDO.

DEFINE INPUT PARAMETER cur-flag-list                    AS CHARACTER    NO-UNDO.

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/

IF cur-start-date = cur-end-date THEN

    DO:

    RUN add-qrymgr-query-buffer-search-spec-date(   cur-query-handle,

                                                    cur-buffer-num,
                                                    cur-field-name, "=",    cur-start-date,

                                                    cur-flag-list).

    END.

ELSE

    DO:

    RUN add-qrymgr-query-buffer-search-spec-date(   cur-query-handle,

                                                    cur-buffer-num,
                                                    cur-field-name, ">=",   cur-start-date,

                                                    cur-flag-list).

    RUN add-qrymgr-query-buffer-search-spec-date(   cur-query-handle,

                                                    cur-buffer-num,
                                                    cur-field-name, "<=",   cur-end-date,

                                                    cur-flag-list).
    END.

&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE add-qrymgr-query-buffer-search-spec-decimal-range {&super-header}:
/****************************************************************************
PURPOSE:

    Add a decimal range to a query buffer spec

PARAMETERS:

    cur-query-handle        Handle to the query

    cur-buffer-num          Buffer position in the query

    cur-field-name          Name of the field to range match against

    cur-start-dec           Starting decimal range.     Ignored if this is "0"

    cur-end-dec             End of the decimal range.   Ignored if this is "0"

                            If cur-start-dec = cur-end-dec then an "=" is used.

    cur-flag-list           Optional flags to control behavior.

                            Flag            Description
                            --------------- ---------------------------------
                            "mandatory"     Don't ignore "0" parameter values

*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE    NO-UNDO.
DEFINE INPUT PARAMETER cur-buffer-num                   AS INTEGER   NO-UNDO.

DEFINE INPUT PARAMETER cur-field-name                   AS CHARACTER NO-UNDO.

DEFINE INPUT PARAMETER cur-start-dec                    AS DECIMAL   NO-UNDO.
DEFINE INPUT PARAMETER cur-end-dec                      AS DECIMAL   NO-UNDO.

DEFINE INPUT PARAMETER cur-flag-list                    AS CHARACTER NO-UNDO.

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Variables */

DEFINE VARIABLE is-mandatory                            AS LOGICAL   NO-UNDO.

/****************************************************************************/

ASSIGN
    is-mandatory = LOOKUP("mandatory", cur-flag-list) > 0
    .

/****************************************************************************/

DO TRANSACTION: /* Transaction */

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

FIND tt-query-buffer
    WHERE tt-query-buffer.qry-handle    = cur-query-handle  AND
          tt-query-buffer.buffer-num    = cur-buffer-num
    EXCLUSIVE-LOCK
    NO-ERROR.

IF NOT AVAILABLE tt-query-buffer    OR      /* No matching query        */
   (cur-start-dec = 0.0 AND                 /* Range not used           */
    cur-end-dec   = 0.0 AND                 /* Range not used           */
    NOT is-mandatory)               THEN    /* Not a mandatory query    */

    LEAVE.

IF cur-start-dec = cur-end-dec THEN
    DO:
    RUN add-query-condition(INPUT-OUTPUT tt-query-buffer.search-spec,
                            cur-buffer-num,
                            cur-field-name,
                            "=",
                            STRING(cur-start-dec)).

    END.

ELSE

    DO:

    IF cur-start-dec <> 0   OR
       is-mandatory         THEN

        RUN add-query-condition(INPUT-OUTPUT tt-query-buffer.search-spec,
                                cur-buffer-num,
                                cur-field-name,
                                ">=",
                                STRING(cur-start-dec)).


    IF cur-end-dec <> 0   OR
       is-mandatory         THEN

        RUN add-query-condition(INPUT-OUTPUT tt-query-buffer.search-spec,
                                cur-buffer-num,
                                cur-field-name,
                                "<=",
                                STRING(cur-end-dec)).
    END.

RUN set-tt-query-join-search-spec(BUFFER tt-query-buffer).

FIND CURRENT tt-query-buffer
    NO-LOCK.

END.            /* Transaction */

&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE add-qrymgr-query-buffer-search-spec-integer-range {&super-header}:
/****************************************************************************
PURPOSE:

    Add a integer range to a query buffer spec

PARAMETERS:

    cur-query-handle        Handle to the query

    cur-buffer-num          Buffer position in the query

    cur-field-name          Name of the field to range match against

    cur-start-int           Starting integer range.     Ignored if this is "0"

    cur-end-int             End of the integer range.   Ignored if this is "0"

                            If cur-start-int = cur-end-int then an "=" is used.

    cur-flag-list           Optional flags to control behavior.

                            Flag            Description
                            --------------- ---------------------------------
                            "mandatory"     Don't ignore "0"

*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE    NO-UNDO.
DEFINE INPUT PARAMETER cur-buffer-num                   AS INTEGER   NO-UNDO.
DEFINE INPUT PARAMETER cur-field-name                   AS CHARACTER NO-UNDO.

DEFINE INPUT PARAMETER cur-start-int                    AS INTEGER   NO-UNDO.
DEFINE INPUT PARAMETER cur-end-int                      AS INTEGER   NO-UNDO.

DEFINE INPUT PARAMETER cur-flag-list                    AS CHARACTER NO-UNDO.

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Variables */

DEFINE VARIABLE is-mandatory                            AS LOGICAL   NO-UNDO.

/****************************************************************************/

ASSIGN
    is-mandatory = LOOKUP("mandatory", cur-flag-list) > 0
    .

/****************************************************************************/

DO TRANSACTION: /* Transaction */

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

FIND tt-query-buffer
    WHERE tt-query-buffer.qry-handle    = cur-query-handle  AND
          tt-query-buffer.buffer-num    = cur-buffer-num
    EXCLUSIVE-LOCK
    NO-ERROR.

IF NOT AVAILABLE tt-query-buffer    OR      /* No matching query        */
   (cur-start-int = 0 AND                   /* Range not used           */
    cur-end-int   = 0 AND                   /* Range not used           */
    NOT is-mandatory)               THEN    /* Not a mandatory query    */

    LEAVE.

IF cur-start-int = cur-end-int THEN

    DO:

    RUN add-query-condition(    INPUT-OUTPUT tt-query-buffer.search-spec,
                                cur-buffer-num, cur-field-name, "=",    STRING(cur-start-int)).

    END.

ELSE

    DO:

    IF cur-start-int <> 0   OR
       is-mandatory         THEN

        RUN add-query-condition(INPUT-OUTPUT tt-query-buffer.search-spec,
                                cur-buffer-num, cur-field-name, ">=",   STRING(cur-start-int)).

    IF cur-end-int <> 0   OR
       is-mandatory         THEN

        RUN add-query-condition(INPUT-OUTPUT tt-query-buffer.search-spec,
                                cur-buffer-num, cur-field-name, "<=",   STRING(cur-end-int)).

    END.

RUN set-tt-query-join-search-spec(BUFFER tt-query-buffer).

FIND CURRENT tt-query-buffer
    NO-LOCK.

END.            /* Transaction */

&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE add-qrymgr-query-buffer-search-spec-character-range   {&super-header}:
/****************************************************************************
PURPOSE:

    Add a character range to a query buffer spec

PARAMETERS:

    cur-query-handle        Handle to the query

    cur-buffer-num          Buffer position in the query

    cur-field-name          Name of the field to range match against

    cur-start-chr           Starting character range.     Ignored if this is ""

    cur-end-chr             End of the character range.   Ignored if this is ""

                            If cur-start-chr = cur-end-chr then an "=" is used.

    cur-flag-list           Optional flags to control behavior.

                            Flag            Description
                            --------------- ---------------------------------
                            "mandatory"     Don't ignore ""

*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE    NO-UNDO.
DEFINE INPUT PARAMETER cur-buffer-num                   AS INTEGER   NO-UNDO.
DEFINE INPUT PARAMETER cur-field-name                   AS CHARACTER NO-UNDO.

DEFINE INPUT PARAMETER cur-start-chr                    AS CHARACTER NO-UNDO.
DEFINE INPUT PARAMETER cur-end-chr                      AS CHARACTER NO-UNDO.

DEFINE INPUT PARAMETER cur-flag-list                    AS CHARACTER NO-UNDO.

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Variables */

DEFINE VARIABLE is-mandatory                            AS LOGICAL   NO-UNDO.

/****************************************************************************/

ASSIGN
    is-mandatory = LOOKUP("mandatory", cur-flag-list) > 0
    .

/****************************************************************************/

DO TRANSACTION: /* Transaction */

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

FIND tt-query-buffer
    WHERE tt-query-buffer.qry-handle    = cur-query-handle  AND
          tt-query-buffer.buffer-num    = cur-buffer-num
    EXCLUSIVE-LOCK
    NO-ERROR.

IF NOT AVAILABLE tt-query-buffer    OR      /* No matching query        */
   (cur-start-chr   = "" AND                /* Range not used           */
    cur-end-chr     = "" AND                /* Range not used           */
    NOT is-mandatory)               THEN    /* Not a mandatory query    */

    LEAVE.

IF cur-start-chr = cur-end-chr THEN
    DO:
    RUN add-query-condition(INPUT-OUTPUT tt-query-buffer.search-spec,
                            cur-buffer-num,
                            cur-field-name,
                            "=",
                            get-qrymgr-quoted-string(cur-start-chr)).

    END.

ELSE
    DO:

    IF cur-start-chr <> ""   OR
       is-mandatory         THEN

        RUN add-query-condition(INPUT-OUTPUT tt-query-buffer.search-spec,
                                cur-buffer-num,
                                cur-field-name,
                                ">=",
                                get-qrymgr-quoted-string(cur-start-chr)).


    IF cur-end-chr <> ""    OR
       is-mandatory         THEN

        RUN add-query-condition(INPUT-OUTPUT tt-query-buffer.search-spec,
                                cur-buffer-num,
                                cur-field-name,
                                "<=",
                                get-qrymgr-quoted-string(cur-end-chr)).
    END.

RUN set-tt-query-join-search-spec(BUFFER tt-query-buffer).

FIND CURRENT tt-query-buffer
    NO-LOCK.

END.            /* Transaction */

&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE add-qrymgr-query-buffer-search-spec-numeric       {&super-header}:
/****************************************************************************
PURPOSE:

    Add a numeric equivalance filter to a query buffer spec from a
    numeric-string

PARAMETERS:

    cur-query-handle        Handle to the query

    cur-buffer-num          Buffer position in the query

    cur-field-name          Name of the field in the buffer to compare against

    cur-condition           Boolean matching condition to use
                            (ie, =, <, >, <=, >=, etc.).

    cur-numeric-chr         Number in string format. Ignored if this is ""

    cur-flag-list           Optional flags to control behavior.

                            Flag            Description
                            --------------- ---------------------------------
                            Not used at this time

*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE    NO-UNDO.
DEFINE INPUT PARAMETER cur-buffer-num                   AS INTEGER   NO-UNDO.
DEFINE INPUT PARAMETER cur-field-name                   AS CHARACTER NO-UNDO.

DEFINE INPUT PARAMETER cur-condition                    AS CHARACTER NO-UNDO.

DEFINE INPUT PARAMETER cur-numeric-chr                  AS CHARACTER NO-UNDO.

DEFINE INPUT PARAMETER cur-flag-list                    AS CHARACTER NO-UNDO.

/****************************************************************************/
&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/

DO TRANSACTION: /* Transaction */

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

FIND tt-query-buffer
    WHERE tt-query-buffer.qry-handle    = cur-query-handle  AND
          tt-query-buffer.buffer-num    = cur-buffer-num
    EXCLUSIVE-LOCK
    NO-ERROR.

IF NOT AVAILABLE tt-query-buffer    OR      /* No matching query        */
   cur-numeric-chr = ""             THEN    /* No value to match filter */

    LEAVE.

RUN add-query-condition(INPUT-OUTPUT tt-query-buffer.search-spec,
                        cur-buffer-num,
                        cur-field-name,
                        cur-condition,
                        cur-numeric-chr).

RUN set-tt-query-join-search-spec(BUFFER tt-query-buffer).

FIND CURRENT tt-query-buffer
    NO-LOCK.

END.            /* Transaction */

&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE add-qrymgr-query-buffer-search-spec-decimal       {&super-header}:
/****************************************************************************
PURPOSE:

    Add a decimal filter to a query buffer spec from a boolean matching condition

PARAMETERS:

    cur-query-handle        Handle to the query

    cur-buffer-num          Buffer position in the query

    cur-field-name          Name of the field in the buffer to compare against

    cur-condition           Boolean matching condition to use
                            (ie, =, <, >, <=, >=, etc.).

    cur-decimal             Decimal value to do the search for

    cur-flag-list           Optional flags to control behavior.

                            Flag            Description
                            --------------- ---------------------------------
                            "mandatory"     Search for values of 0 and ?

*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE       NO-UNDO.
DEFINE INPUT PARAMETER cur-buffer-num                   AS INTEGER      NO-UNDO.
DEFINE INPUT PARAMETER cur-field-name                   AS CHARACTER    NO-UNDO.

DEFINE INPUT PARAMETER cur-condition                    AS CHARACTER    NO-UNDO.

DEFINE INPUT PARAMETER cur-decimal                      AS DECIMAL      NO-UNDO.

DEFINE INPUT PARAMETER cur-flag-list                    AS CHARACTER    NO-UNDO.

/****************************************************************************/
&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Variables                      */

DEFINE VARIABLE is-mandatory                            AS LOGICAL      NO-UNDO.

/****************************************************************************/

DO TRANSACTION: /* Transaction */

    /* Set flags            */

ASSIGN
    is-mandatory = LOOKUP("mandatory", cur-flag-list) > 0
    .

    /* look for the query   */

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

FIND tt-query-buffer
    WHERE tt-query-buffer.qry-handle    = cur-query-handle  AND
          tt-query-buffer.buffer-num    = cur-buffer-num
    EXCLUSIVE-LOCK
    NO-ERROR.

    /* Check conditions     */

IF NOT AVAILABLE tt-query-buffer            OR          /* No matching query            */
   ((cur-decimal = 0     OR
     cur-decimal = ?) AND NOT is-mandatory) THEN        /* No value to match filter     */

    LEAVE.

RUN add-query-condition(INPUT-OUTPUT tt-query-buffer.search-spec,
                        cur-buffer-num,
                        cur-field-name,
                        cur-condition,
                        STRING(cur-decimal)).

RUN set-tt-query-join-search-spec(BUFFER tt-query-buffer).

FIND CURRENT tt-query-buffer
    NO-LOCK.

END.            /* Transaction */

&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE add-qrymgr-query-buffer-search-spec-integer       {&super-header}:
/****************************************************************************
PURPOSE:

    Add a integer filter to a query buffer spec from a boolean matching condition

PARAMETERS:

    cur-query-handle        Handle to the query

    cur-buffer-num          Buffer position in the query

    cur-field-name          Name of the field in the buffer to compare against

    cur-condition           Boolean matching condition to use
                            (ie, =, <, >, <=, >=, etc.).

    cur-integer             Decimal value to do the search for

    cur-flag-list           Optional flags to control behavior.

                            Flag            Description
                            --------------- ---------------------------------
                            "mandatory"     Search for values of 0 and ?

*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE       NO-UNDO.
DEFINE INPUT PARAMETER cur-buffer-num                   AS INTEGER      NO-UNDO.
DEFINE INPUT PARAMETER cur-field-name                   AS CHARACTER    NO-UNDO.

DEFINE INPUT PARAMETER cur-condition                    AS CHARACTER    NO-UNDO.

DEFINE INPUT PARAMETER cur-integer                      AS INTEGER      NO-UNDO.

DEFINE INPUT PARAMETER cur-flag-list                    AS CHARACTER    NO-UNDO.

/****************************************************************************/
&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Variables                      */

DEFINE VARIABLE is-mandatory                            AS LOGICAL      NO-UNDO.

/****************************************************************************/

DO TRANSACTION: /* Transaction */

    /* Set flags            */

ASSIGN
    is-mandatory = LOOKUP("mandatory", cur-flag-list) > 0
    .

    /* look for the query   */

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

FIND tt-query-buffer
    WHERE tt-query-buffer.qry-handle    = cur-query-handle  AND
          tt-query-buffer.buffer-num    = cur-buffer-num
    EXCLUSIVE-LOCK
    NO-ERROR.

    /* Check conditions     */

IF NOT AVAILABLE tt-query-buffer            OR          /* No matching query            */
   ((cur-integer = 0     OR
     cur-integer = ?) AND NOT is-mandatory) THEN        /* No value to match filter     */

    LEAVE.

RUN add-query-condition(INPUT-OUTPUT tt-query-buffer.search-spec,
                        cur-buffer-num,
                        cur-field-name,
                        cur-condition,
                        STRING(cur-integer)).

RUN set-tt-query-join-search-spec(BUFFER tt-query-buffer).

FIND CURRENT tt-query-buffer
    NO-LOCK.

END.            /* Transaction */

&ENDIF 
END PROCEDURE.

/****************************************************************************/
PROCEDURE add-qrymgr-query-buffer-search-spec-logical       {&super-header}:
/****************************************************************************
PURPOSE:

    Add a logical filter to a query buffer spec from a boolean matching condition

PARAMETERS:

    cur-query-handle        Handle to the query

    cur-buffer-num          Buffer position in the query

    cur-field-name          Name of the field in the buffer to compare against

    cur-condition           Boolean matching condition to use
                            (ie, =, <, >, <=, >=, etc.).

    cur-logical             Logical value to match.

    cur-flag-list           Optional flags to control behavior.

                            Flag            Description
                            --------------- ---------------------------------
                            Not used at this time

*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE    NO-UNDO.
DEFINE INPUT PARAMETER cur-buffer-num                   AS INTEGER   NO-UNDO.
DEFINE INPUT PARAMETER cur-field-name                   AS CHARACTER NO-UNDO.

DEFINE INPUT PARAMETER cur-condition                    AS CHARACTER NO-UNDO.

DEFINE INPUT PARAMETER cur-logical                      AS CHARACTER NO-UNDO.

DEFINE INPUT PARAMETER cur-flag-list                    AS CHARACTER NO-UNDO.

/****************************************************************************/
&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Variables                      */

DEFINE VARIABLE is-mandatory                            AS LOGICAL      NO-UNDO.

/****************************************************************************/

DO TRANSACTION: /* Transaction */

    /* Set flags            */

ASSIGN
    is-mandatory = LOOKUP("mandatory", cur-flag-list) > 0
    .


RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

FIND tt-query-buffer
    WHERE tt-query-buffer.qry-handle    = cur-query-handle  AND
          tt-query-buffer.buffer-num    = cur-buffer-num
    EXCLUSIVE-LOCK
    NO-ERROR.

IF NOT AVAILABLE tt-query-buffer            OR          /* No matching query                */
   (cur-logical = ?  AND NOT is-mandatory)  THEN        /* No value to match filter         */

    LEAVE.

RUN add-query-condition(INPUT-OUTPUT tt-query-buffer.search-spec,
                        cur-buffer-num,
                        cur-field-name,
                        cur-condition,
                        STRING(cur-logical)).

RUN set-tt-query-join-search-spec(BUFFER tt-query-buffer).

FIND CURRENT tt-query-buffer
    NO-LOCK.

END.            /* Transaction */

&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE add-qrymgr-query-buffer-search-spec-character     {&super-header}:
/****************************************************************************
PURPOSE:

    Add a character search to a query buffer spec. Optionally addes a MATCH
    specification if "can-match" is specified in the flag list and a "match"
    character is specified (ie a "*" or ".") in the search string.

PARAMETERS:

    cur-query-handle        Handle to the query

    cur-buffer-num          Numeric buffer position in the query

    cur-field-name          Name of the field in the buffer to compare against

    cur-condition           Boolean matching condition to use
                            (ie, =, <, >, <=, >=, etc.). This is replaced with
                            MATCH if "can-match" is specified and the cur-chr
                            contains "*" or "." characters.

    cur-chr                 String to compare for. Ignored if this is "" and
                            "mandatory" is not specified in the flag list.

    cur-flag-list           Optional flags that affect the search specification
                            construction.

                            Flag            Description
                            --------------- ---------------------------------
                            "mandatory"     Don't ignore ""

                            "can-match"     Use MATCH instead of the logical
                                            condition if "*" or "." is in the
                                            cur-chr string

                            "assume-match"  Use MATCH "*chars*" unless
                                            a "*" or "." is specified in the
                                            search string

*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE    NO-UNDO.
DEFINE INPUT PARAMETER cur-buffer-num                   AS INTEGER   NO-UNDO.
DEFINE INPUT PARAMETER cur-field-name                   AS CHARACTER NO-UNDO.

DEFINE INPUT PARAMETER cur-condition                    AS CHARACTER NO-UNDO.

DEFINE INPUT PARAMETER cur-chr                          AS CHARACTER NO-UNDO.

DEFINE INPUT PARAMETER cur-flag-list                    AS CHARACTER NO-UNDO.

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Variables */

DEFINE VARIABLE is-mandatory                            AS LOGICAL   NO-UNDO.
DEFINE VARIABLE is-can-match                            AS LOGICAL   NO-UNDO.
DEFINE VARIABLE is-assume-match                         AS LOGICAL   NO-UNDO.

/****************************************************************************/

ASSIGN
    is-mandatory    = LOOKUP("mandatory",       cur-flag-list) > 0
    is-can-match    = LOOKUP("can-match",       cur-flag-list) > 0
    is-assume-match = LOOKUP("assume-match",    cur-flag-list) > 0
    .

/****************************************************************************/

DO TRANSACTION: /* Transaction */

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

FIND tt-query-buffer
    WHERE tt-query-buffer.qry-handle    = cur-query-handle  AND
          tt-query-buffer.buffer-num    = cur-buffer-num
    EXCLUSIVE-LOCK
    NO-ERROR.

IF NOT AVAILABLE tt-query-buffer        OR              /* No matching query                */
   (cur-chr = "" AND NOT is-mandatory)  THEN            /* Nothing to look for              */

    LEAVE.

IF is-can-match     OR
   is-assume-match  THEN
    DO:

    IF REPLACE(cur-chr, "*", "") <> cur-chr OR
       REPLACE(cur-chr, ".", "") <> cur-chr THEN

        ASSIGN
            cur-condition   = "MATCHES"         WHEN is-can-match
            .

    ELSE

        IF is-assume-match THEN
            ASSIGN
                cur-chr = "*" + cur-chr + "*"   WHEN is-assume-match
                .

    END.

RUN add-query-condition(INPUT-OUTPUT tt-query-buffer.search-spec,
                        cur-buffer-num,
                        cur-field-name,
                        cur-condition,
                        get-qrymgr-quoted-string(cur-chr)).

RUN set-tt-query-join-search-spec(BUFFER tt-query-buffer).

FIND CURRENT tt-query-buffer
    NO-LOCK.

END.            /* Transaction */

&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE add-qrymgr-query-buffer-search-spec-date          {&super-header}:
/****************************************************************************
PURPOSE:

    Add a date search to a query buffer spec.

PARAMETERS:

    cur-query-handle        Handle to the query

    cur-buffer-num          Numeric buffer position in the query

    cur-field-name          Name of the field in the buffer to compare against

    cur-condition           Boolean matching condition to use
                            (ie, =, <, >, <=, >=, etc.). 

    cur-date                Date to compare for. Ignored if this is ? and
                            "mandatory" is not specified in the flag list.

    cur-flag-list           Optional flags that affect the search specification
                            construction.

                            Flag            Description
                            --------------- ---------------------------------
                            "mandatory"     Don't ignore ? cur-date value

*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-query-handle             AS HANDLE       NO-UNDO.
DEFINE INPUT PARAMETER cur-buffer-num               AS INTEGER      NO-UNDO.
DEFINE INPUT PARAMETER cur-field-name               AS CHARACTER    NO-UNDO.

DEFINE INPUT PARAMETER cur-condition                AS CHARACTER    NO-UNDO.

DEFINE INPUT PARAMETER cur-date                     AS DATE         NO-UNDO.

DEFINE INPUT PARAMETER cur-flag-list                AS CHARACTER    NO-UNDO.

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Variables */

DEFINE VARIABLE is-mandatory                            AS LOGICAL   NO-UNDO.

/****************************************************************************/

ASSIGN
    is-mandatory    = LOOKUP("mandatory", cur-flag-list) > 0
    .

/****************************************************************************/

DO TRANSACTION: /* Transaction */

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

FIND tt-query-buffer
    WHERE tt-query-buffer.qry-handle    = cur-query-handle  AND
          tt-query-buffer.buffer-num    = cur-buffer-num
    EXCLUSIVE-LOCK
    NO-ERROR.

IF NOT AVAILABLE tt-query-buffer        OR      /* No matching query    */
   (cur-date = ? AND NOT is-mandatory)  THEN    /* Nothing to look for  */

    LEAVE.

RUN add-query-condition(INPUT-OUTPUT tt-query-buffer.search-spec,
                        cur-buffer-num,
                        cur-field-name,
                        cur-condition,

                        IF cur-date = ?
                            THEN "?"
                            ELSE "DATE(" + STRING(INTEGER(cur-date)) + ")").

RUN set-tt-query-join-search-spec(BUFFER tt-query-buffer).

FIND CURRENT tt-query-buffer
    NO-LOCK.

END.            /* Transaction */

&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE reset-qrymgr-query-buffer-search-spec             {&super-header}:
/****************************************************************************
PURPOSE:

    Reset the search spec for a given query / buffer combination. This is
    for filtering the queries based on a given buffer, and doesn't affect
    the table join specification for this buffer.

PARAMETERS:

    cur-query-handle    Handle the query in question

    cur-buffer-num      Buffer number to clear the search spec for

*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE    NO-UNDO.
DEFINE INPUT PARAMETER cur-buffer-num                   AS INTEGER   NO-UNDO.

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Buffers */

DEFINE BUFFER tt-query-buffer                           FOR tt-query-buffer.

/****************************************************************************/

DO TRANSACTION: /* Transaction */

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

FIND tt-query-buffer
    WHERE tt-query-buffer.qry-handle    = cur-query-handle AND
          tt-query-buffer.buffer-num    = cur-buffer-num
    EXCLUSIVE-LOCK
    NO-ERROR.

IF NOT AVAILABLE tt-query-buffer THEN
    LEAVE.

ASSIGN
    tt-query-buffer.search-spec = ""
    .

RUN set-tt-query-join-search-spec(BUFFER tt-query-buffer).

FIND CURRENT tt-query-buffer
    NO-LOCK.

END.            /* Transaction */

&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE open-qrymgr-query                             {&super-header}:
/****************************************************************************
PURPOSE:

    Open a query with the supplied search / sort specification.

    If the query handle points to a static query that isn't under
    management by this code, then import it first and then open it
    with the supplied specification

PARAMETERS:

    cur-query-handle    Handle to the query to open

    cur-search-spec     Search specification to open the query with.
                        The format to use is:

                        FOR EACH &1
                            WHERE &1.field = condition,

                            EACH &2
                                WHERE &2.field  = &1.field AND
                                      &2.field2 = value

                        The function get-qrymgr-quoted-string() should be
                        used to properly encode string constant values
                        when constructing a search spec

                        NOTE: A blank ("") search specification indicates that
                        the search spec is to be generated from the associated
                        join / search specs in the tt-query-buffer table

    cur-sort-order      Sort order for the result list. This is a comma-delimited
                        list.

                        For example:

                            &1.field1,&2.field2,&3.field3

                        or

                            &1.field1,&1.field2

                        or

                            &1.field1 DESCENDING,&1.field2


    cur-flag-list       Optional flags to control behavior.

                        Flag                    Description
                        ---------------         -------------------------------------
                        "reposition"            This is a scrolling query, so forward-only = false
                                                (9.1D and following versions only)

                                                This is intended as an override for the default
                                                setting applied by import-qrymgr-query().

                        "keep-position"         Save the current buffer ROWID position(s)
                                                and then try to reposition to those same
                                                ROWIDs when the query's opened.

                        "invert-sort"           Invert the specified sort order

                        "invert-buffer-order"   Invert the order of the buffers in the
                                                query - ie

                                                    FOR EACH buf1,
                                                        EACH buf2
                                                            WHERE buf2.field = buf1.field

                                                would become

                                                    FOR EACH buf2
                                                        WHERE buf2.field = buf1.field,

                                                        EACH buf1

*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE    NO-UNDO.
DEFINE INPUT PARAMETER cur-search-spec                  AS CHARACTER NO-UNDO.
DEFINE INPUT PARAMETER cur-sort-order                   AS CHARACTER NO-UNDO.

DEFINE INPUT PARAMETER cur-flag-list                    AS CHARACTER NO-UNDO.   /* Reserved for possible future use */

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Variables */

DEFINE VARIABLE cur-buf-num-string                      AS CHARACTER NO-UNDO.
DEFINE VARIABLE cur-entry-value                         AS CHARACTER NO-UNDO.
DEFINE VARIABLE cur-entry-num                           AS INTEGER   NO-UNDO.

DEFINE VARIABLE is-keep-position                        AS LOGICAL   NO-UNDO.
DEFINE VARIABLE is-invert-sort-order                    AS LOGICAL   NO-UNDO.

DEFINE VARIABLE cur-num-entries                         AS INTEGER   NO-UNDO.
DEFINE VARIABLE is-reposition                           AS LOGICAL   NO-UNDO.

/****************************************************************************/

ASSIGN
    is-keep-position        = LOOKUP("keep-position",   cur-flag-list) > 0
    is-invert-sort-order    = LOOKUP("invert-sort",     cur-flag-list) > 0

    is-reposition           = LOOKUP("reposition",      cur-flag-list) > 0
    .

/****************************************************************************/

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).                              /* Make sure this is a query handle             */

IF NOT CAN-FIND(tt-query                                                        /* If this is an externally defined query,      */
            WHERE tt-query.qry-handle = cur-query-handle) THEN

    RUN import-qrymgr-query(cur-query-handle, "").                              /* then import it's information automatically   */

/****************************************************************************/
/* IF there's no specified spec,            */
/* then derive it from the buffer records   */

IF cur-search-spec = "" THEN
    DO:

    FOR EACH tt-query-buffer
        WHERE tt-query-buffer.qry-handle = cur-query-handle

        NO-LOCK

        BREAK BY tt-query-buffer.qry-handle
              BY tt-query-buffer.buffer-num:

        IF FIRST-OF(tt-query-buffer.qry-handle) THEN

            ASSIGN
                cur-search-spec = "FOR "
                .

        ELSE

            ASSIGN
                cur-search-spec = cur-search-spec + ","
                .

        ASSIGN
            cur-search-spec =   cur-search-spec                                 + " " +     /* Prior search levels                      */

                                (IF tt-query-buffer.join-record-set > ""                    /* Record join set size                     */
                                    THEN  tt-query-buffer.join-record-set                   /* as specified                             */
                                    ELSE  "EACH")                               + " " +     /* or "Each" as a default                   */
                                                                                            /* FIRST,EACH, or LAST                      */
                                "&" + TRIM(STRING(tt-query-buffer.buffer-num))  + " " +

                                (IF tt-query-buffer.join-type > "" THEN                     /* Optional outer-join or left outer-join   */
                                    TRIM(tt-query-buffer.join-type) + " "
                                    ELSE "")                                    +

                                (IF tt-query-buffer.join-search-spec <> ""                  /* any "where" condition?                   */
                                    THEN "WHERE " + tt-query-buffer.join-search-spec
                                    ELSE "")
            .

    END.

    END.

/****************************************************************************/
/* Take the field names and put them into the query string */

FOR EACH tt-query-buffer

    WHERE tt-query-buffer.qry-handle = cur-query-handle

    BY tt-query-buffer.qry-handle DESCENDING            /* MUST BE DONE IN DESCENDING NUMERIC ORDER!    */
    BY tt-query-buffer.buffer-num DESCENDING:           /* Don't want &11 being hit by replacement of   */
                                                        /* &1 when its being processed                  */

        /* REPLACE is used instead of SUBSTITUTE        */
        /* since SUBSTITUTE can only handle 9 values,   */
        /* but queries can have up to 18 buffers.       */

    ASSIGN
        cur-search-spec = REPLACE(  cur-search-spec,
                                    "&" + TRIM(STRING(tt-query-buffer.buffer-num)),
                                    tt-query-buffer.db-table-name)

        cur-sort-order  = REPLACE(  cur-sort-order,
                                    "&" + TRIM(STRING(tt-query-buffer.buffer-num)),
                                    "BY " + tt-query-buffer.db-table-name)
        .

END.

/****************************************************************************/
/* Invert the sort order if required    */

IF is-invert-sort-order THEN
    DO:

    ASSIGN
        cur-num-entries = NUM-ENTRIES(cur-sort-order)
        .

    DO cur-entry-num = 1 TO cur-num-entries:

        ASSIGN
            cur-entry-value = ENTRY(cur-entry-num, cur-sort-order)
            .


       ENTRY(cur-entry-num, cur-sort-order) = IF LOOKUP(ENTRY(NUM-ENTRIES(cur-entry-value, " "), cur-entry-value, " "), "descending,desc") > 0
                                                THEN REPLACE(ENTRY(cur-entry-num, cur-sort-order), "descending", "")
                                                ELSE cur-entry-value + " descending".


    END.

    END.    /* IF is-invert-sort-order  */

/****************************************************************************/
/* Replace all the commas with spaces   */
/* in the sort order                    */

ASSIGN
    cur-sort-order = REPLACE(cur-sort-order, ",", " ")
    .

/****************************************************************************/
/* Save the query specification     */

DO TRANSACTION: /* Transaction */

FIND tt-query
    WHERE tt-query.qry-handle = cur-query-handle
    EXCLUSIVE-LOCK
    NO-ERROR.

ASSIGN
    tt-query.qry-spec = cur-search-spec + " " +
                        cur-sort-order
    .

FIND CURRENT tt-query
    NO-LOCK.

END.            /* Transaction */

/****************************************************************************/
/* Open the query                           */

    /* Save the current query position      */

IF is-keep-position THEN
    RUN set-qrymgr-query-position(  cur-query-handle,
                                    cur-local-name,
                                    "save").

    /* Set the browse REFRESHABLE = NO      */

RUN control-browse(BUFFER tt-query, "disable").

/****************************************************************************/
/* Change the query if "forward-only" gets  */
/* changed                                  */

&IF KEYWORD-ALL( "FORWARD-ONLY" ) NE ? &THEN

IF is-reposition THEN
    DO:

    IF cur-query-handle:FORWARD-ONLY <> NO THEN
        DO:

        IF cur-query-handle:IS-OPEN THEN
            cur-query-handle:QUERY-CLOSE().

        ASSIGN
            cur-query-handle:FORWARD-ONLY = NO
            .

        END.

    END.

ELSE    /* The caller didn't specify "reposition",  */
        /* so put it back to it's original setting  */

    IF cur-query-handle:FORWARD-ONLY <> tt-query.is-forward-only THEN
        DO:

        IF cur-query-handle:IS-OPEN THEN
            cur-query-handle:QUERY-CLOSE().

        ASSIGN
            cur-query-handle:FORWARD-ONLY = tt-query.is-forward-only
            .

        END.

&ENDIF

/****************************************************************************/

    /* Prepare and open the query               */

tt-query.qry-handle:QUERY-PREPARE(tt-query.qry-spec).

tt-query.qry-handle:QUERY-OPEN().

    /* Try to return to the original record(s)  */

IF is-keep-position THEN

    RUN set-qrymgr-query-position(  cur-query-handle,
                                    cur-local-name,
                                    "recall").

ELSE

    IF tt-query.is-browse-control   THEN
        tt-query.qry-handle:GET-FIRST(NO-LOCK).

    /* Set the browse REFRESHABLE = YES         */
    /* and set the browse position              */

RUN control-browse(BUFFER tt-query, "enable").

&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE reopen-qrymgr-query                               {&super-header}:
/****************************************************************************
PURPOSE:

    Reopen a query, assuming we're going to use the last used search spec

PARAMETERS:

    cur-query-handle    Which managed query to reopen

    cur-flag-list       Optional flags to control behavior.

                        Flag            Description
                        --------------- -------------------------------------
                        "keep-position" Remember the current position of all
                                        the associated buffers and attempt to
                                        go back to the same location in the
                                        query after it's been re-opened.

*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE    NO-UNDO.
DEFINE INPUT PARAMETER cur-flag-list                    AS CHARACTER NO-UNDO.

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Variables */

DEFINE VARIABLE is-keep-position                        AS LOGICAL   NO-UNDO.

/****************************************************************************/
/* Local Buffers */

DEFINE BUFFER tt-query                                  FOR tt-query.

/****************************************************************************/

ASSIGN
    is-keep-position = LOOKUP("keep-position", cur-flag-list) > 0
    .

/****************************************************************************/

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

FIND tt-query
    WHERE tt-query.qry-handle = cur-query-handle
    NO-LOCK
    NO-ERROR.

IF NOT AVAILABLE tt-query               OR
   NOT VALID-HANDLE(tt-query.qry-handle) THEN

    LEAVE.

RUN control-browse(BUFFER tt-query, "disable").                     /* Stop the browse from flashing            */

IF is-keep-position THEN                                            /* If we want to end up back where we were  */
    RUN set-qrymgr-query-position(  cur-query-handle,               /* Save the current position                */
                                    cur-local-name,
                                    "save").

tt-query.qry-handle:QUERY-OPEN().

IF is-keep-position THEN                                            /* If we want to end up back where we were  */
    RUN set-qrymgr-query-position(  cur-query-handle,               /* return to the current position           */
                                    cur-local-name,
                                    "recall").

RUN control-browse(BUFFER tt-query, "enable").                      /* Update the browse contents on the screen */

&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE set-qrymgr-query-position                         {&super-header}:
/****************************************************************************
PURPOSE:

    Save or recall buffer positions for a given query

PARAMETERS:

    cur-query-handle    Handle to the query to do this for

    cur-position-name   Name of the position this corresponds to. This can be
                        used to create multiple "save" positions and restore to
                        them as required.

                        NOTE: The position name

                                    THIS-PROCEDURE:FILE-NAME + STRING(THIS-PROCEDURE)

                              is reserved for internal use..

    cur-command         What to do. Legal commands are

                        Command         Description
                        --------------- -------------------------------------
                        "delete"        Delete a named position for this query
                        
                        "recall"        Recall the current buffer positions
                                        from the tt-query-buffer table and
                                        reposition the query there.

                        "save"          Save all the current buffer positions
                                        to the tt-query-buffer table. This will
                                        create a new position record if none
                                        exists already

*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-query-handle             AS HANDLE       NO-UNDO.
DEFINE INPUT PARAMETER cur-position-name            AS CHARACTER    NO-UNDO.
DEFINE INPUT PARAMETER cur-command                  AS CHARACTER    NO-UNDO.

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Variables */

DEFINE VARIABLE cur-cnt                             AS INTEGER   NO-UNDO.

DEFINE VARIABLE is-reposition                       AS LOGICAL   NO-UNDO.

/****************************************************************************/

IF NOT VALID-HANDLE(cur-query-handle) THEN
    LEAVE.

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

FIND tt-query-position
    WHERE tt-query-position.qry-handle      = cur-query-handle  AND
          tt-query-position.position-name   = cur-position-name
    NO-LOCK
    NO-ERROR.

CASE cur-command:

/****************************************************************************/
WHEN "delete"   THEN

    DO TRANSACTION: /* Transaction */

    IF AVAILABLE tt-query-position THEN
        DO:

        FIND CURRENT tt-query-position
            EXCLUSIVE-LOCK.

        DELETE tt-query-position.

        END.

    END.            /* Transaction */

/****************************************************************************/
WHEN "recall"   THEN

    DO:

        /* Make sure the saved position is still there      */

    IF NOT AVAILABLE tt-query-position THEN
        LEAVE.

        /* Assume the reposition didn't work                */

    ASSIGN
        is-reposition = NO
        .

        /* Then reposition the query to the saved position  */

    IF tt-query-position.buffer-positions[1] <> ? THEN

        ASSIGN
            is-reposition = tt-query-position.qry-handle:REPOSITION-TO-ROWID(tt-query-position.buffer-positions)
            NO-ERROR.

    ELSE    /* or not...                                    */

        ASSIGN
            is-reposition = NO
            .

        /* If reposition worked, then                       */
        /* select the focused row                           */

    IF is-reposition THEN

        RUN control-browse(BUFFER tt-query, "select-focused-row").

    ELSE    /* If reposition didn't work then               */
            /* select the first row                         */

        DO:

        tt-query-position.qry-handle:GET-FIRST(NO-LOCK).

        RUN control-browse(BUFFER tt-query, "select-row-1").

        END.

    END.

/****************************************************************************/
WHEN "save"     THEN

    DO TRANSACTION: /* Transaction */

        /* Find the reposition record,          */
        /* or make a new one if it's not there  */

    IF AVAILABLE tt-query-position THEN

        FIND CURRENT tt-query-position
            EXCLUSIVE-LOCK.

    ELSE

        DO:

        CREATE tt-query-position.

        ASSIGN
            tt-query-position.qry-handle    = cur-query-handle
            tt-query-position.position-name = cur-position-name
            .

        END.

        /* Clear the previous saved positions               */

    ASSIGN
        tt-query-position.buffer-positions = ?
        .

        /* Get the buffer rowids and put them into an array */

    DO cur-cnt = 1 TO MAX(18, cur-query-handle:NUM-BUFFERS):

        FIND tt-query-buffer
            WHERE tt-query-buffer.qry-handle = tt-query-position.qry-handle AND
                  tt-query-buffer.buffer-num = cur-cnt
            NO-LOCK
            NO-ERROR.

        IF AVAILABLE tt-query-buffer                AND
           tt-query-buffer.BUFFER-HANDLE:AVAILABLE  THEN

            ASSIGN
                tt-query-position.buffer-positions[cur-cnt] = tt-query-buffer.BUFFER-HANDLE:ROWID
                .

        ELSE

            ASSIGN
                tt-query-position.buffer-positions[cur-cnt] = ?
                .
    END.

    END.
    
END CASE.

&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE move-qrymgr-query-position                        {&super-header}:
/****************************************************************************
PURPOSE:

    Move the query one direction or another in the result list, and stay
    within the list if the operation would've moved the query outside of
    the result list.

    NOTE:   The "prev" direction and "stay-in-list" flag option will not
            work with queries that are defined with forward-only = TRUE.

PARAMETERS:

    cur-query-handle    Handle to the query

    cur-direction       Which direction to move.

                        Value                   Description
                        ---------------         -------------------------------------
                        "next"                  Move to the next entry in the result
                                                list "prev" Move to the previous entry
                                                in the result list

    cur-flag-list       Optional flags t        o control behavior.

                        Flag                    Description
                        ---------------         -------------------------------------
                        "stay-in-list"          If the specified operation would move
                                                the query outside of the result list,
                                                then don't go past the first / last
                                                record(s) in the result list


                        "delete-current-row"    Reposition the query in the desired
                                                direction so that when the current
                                                record is deleted, the query will
                                                still be on a query record

*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE    NO-UNDO.
DEFINE INPUT PARAMETER cur-direction                    AS CHARACTER NO-UNDO.
DEFINE INPUT PARAMETER cur-flag-list                    AS CHARACTER NO-UNDO.

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Variables */

DEFINE VARIABLE is-stay-in-list                         AS LOGICAL   NO-UNDO.
DEFINE VARIABLE is-delete-current-record                AS LOGICAL   NO-UNDO.

/****************************************************************************/

ASSIGN
    is-stay-in-list             = LOOKUP("stay-in-list",            cur-flag-list) > 0
    is-delete-current-record    = LOOKUP("delete-current-record",   cur-flag-list) > 0
    .

IF NOT VALID-HANDLE(cur-query-handle) THEN
    LEAVE.

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

/****************************************************************************/

CASE cur-direction:

/****************************************************************************/
WHEN "next"     THEN

    DO:

    cur-query-handle:GET-NEXT(NO-LOCK).

    IF NOT cur-query-handle:QUERY-OFF-END THEN
        LEAVE.

    IF is-stay-in-list          OR
       is-delete-current-record THEN

        cur-query-handle:GET-LAST(NO-LOCK).

    IF is-delete-current-record THEN
        DO:

        IF cur-query-handle:QUERY-OFF-END THEN
            LEAVE.

        cur-query-handle:GET-PREV(NO-LOCK).         /* Second to last record */

        IF cur-query-handle:QUERY-OFF-END THEN
            DO:
            cur-query-handle:GET-LAST(NO-LOCK).
            LEAVE.
            END.

        END.

    END.

/****************************************************************************/
WHEN "prev"     THEN

    DO:

    cur-query-handle:GET-PREV(NO-LOCK).

    IF NOT cur-query-handle:QUERY-OFF-END THEN
        LEAVE.

    IF is-stay-in-list          OR
       is-delete-current-record THEN

        cur-query-handle:GET-FIRST(NO-LOCK).        /* Back to the original record      */

    IF is-delete-current-record THEN
        DO:

        cur-query-handle:GET-NEXT(NO-LOCK).         /* Second record in result list     */

        IF cur-query-handle:QUERY-OFF-END THEN      /* Off the other end of the range?  */
            cur-query-handle:GET-FIRST(NO-LOCK).    /* back to the initial record       */

        END.

    END.

END CASE.

&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE close-qrymgr-query                                {&super-header}:
/****************************************************************************
PURPOSE:

    Close a query that's not needed any more, if it's against a valid handle

PARAMETERS:

    cur-query-handle        Handle to the query to close

    cur-flag-list           Optional flags to control behavior.

                            Flag            Description
                            --------------- ---------------------------------
                            Not used. For possible future use

*****************************************************************************/
/* Local Parameters         */

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE    NO-UNDO.
DEFINE INPUT PARAMETER cur-flag-list                    AS CHARACTER NO-UNDO.   /* Reserved for possible future use */

/****************************************************************************/

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/

IF VALID-HANDLE(cur-query-handle) THEN
    DO:

    RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */
    cur-query-handle:QUERY-CLOSE().

    END.

&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE delete-qrymgr-query                               {&super-header}:
/****************************************************************************
PURPOSE:

    Delete a managed query and any associated dynamic buffers that may've
    been created when the query was instantiated

PARAMETERS:

    cur-query-handle    Main query to get rid of

*****************************************************************************/
/* Local Parameters         */

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE    NO-UNDO.

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Buffers */

DEFINE BUFFER tt-query      FOR tt-query.

/****************************************************************************/

DO TRANSACTION: /* Transaction */

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).              /* Make sure this is a query handle         */

FIND tt-query                                                   /* Find the query in question               */
    WHERE tt-query.qry-handle = cur-query-handle
    EXCLUSIVE-LOCK
    NO-ERROR.

IF NOT AVAILABLE tt-query THEN                                  /* Not there?                               */
    LEAVE.                                                      /* then get out                             */

RUN close-qrymgr-query(cur-query-handle, "").                   /* Make sure it's closed                    */

IF tt-query.is-created-local THEN                               /* If this code made this query             */
    DELETE OBJECT tt-query.qry-handle.                          /* then nuke the query object               */

DELETE tt-query.                                                /* Delete the tt record                     */

FOR EACH tt-query-buffer                                        /* Go through all the query buffers         */
    WHERE tt-query-buffer.qry-handle = cur-query-handle
    EXCLUSIVE-LOCK:

    IF tt-query-buffer.is-created-local             AND         /* delete our buffer objects                */
       VALID-HANDLE(tt-query-buffer.buffer-handle)  THEN

        DELETE OBJECT tt-query-buffer.buffer-handle.

    DELETE tt-query-buffer.                                     /* then delete the temp-table record        */

END.

FOR EACH tt-query-position                                      /* And all saved query positions            */
    WHERE tt-query-position.qry-handle = cur-query-handle
    EXCLUSIVE-LOCK:

    DELETE tt-query-position.

END.

FOR EACH tt-query-buffer-field-set-value                        /* All associated buffer field value sets   */
    WHERE tt-query-buffer-field-set-value.qry-handle = cur-query-handle
    EXCLUSIVE-LOCK:

    DELETE tt-query-buffer-field-set-value.

END.

END.            /* Transaction */

&ENDIF 
END PROCEDURE.

/****************************************************************************/
FUNCTION get-qrymgr-query-spec
/****************************************************************************
PURPOSE:

    Get the last used search specification for a given query

PARAMETERS:

    cur-query-handle    query to get the query specification of

*****************************************************************************/

    RETURNS CHARACTER

    (INPUT cur-query-handle     AS HANDLE)                  {&super-header}:

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

FIND tt-query
    WHERE tt-query.qry-handle = cur-query-handle
    NO-LOCK
    NO-ERROR.

RETURN(IF AVAILABLE tt-query
        THEN tt-query.qry-spec
        ELSE "").

END FUNCTION.
&ENDIF

/****************************************************************************/
FUNCTION get-qrymgr-buffer-db-table-handle
/****************************************************************************
PURPOSE:

   Get the handle associated with the named buffer for this query that
   this manager knows about and return it

PARAMETERS:

    cur-query-handle    query handle to get the query specification of

    cur-db-table-name   db.table name specification for the buffer

*****************************************************************************/
    RETURNS HANDLE

    (INPUT cur-query-handle     AS HANDLE,
     INPUT cur-db-table-name    AS CHARACTER)               {&super-header}:

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Buffers */

DEFINE BUFFER tt-query-buffer FOR tt-query-buffer.

/****************************************************************************/

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

FIND tt-query-buffer
    WHERE   tt-query-buffer.qry-handle      = cur-query-handle  AND
            tt-query-buffer.db-table-name   = cur-db-table-name
    NO-LOCK
    NO-ERROR.

RETURN(IF AVAILABLE tt-query-buffer
        THEN tt-query-buffer.buffer-handle
        ELSE ?
        ).

END FUNCTION.
&ENDIF

/****************************************************************************/
FUNCTION get-qrymgr-buffer-order-handle
/****************************************************************************
PURPOSE:

    Get the n'th buffer associated with this query and return it

PARAMETERS:

    cur-query-handle    query to get the query specification of

    cur-buffer-num      Positional number of the buffer in the query

*****************************************************************************/
    RETURNS HANDLE

    (INPUT cur-query-handle     AS HANDLE,
     INPUT cur-buffer-num       AS INTEGER)                 {&super-header}:

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Buffers */

DEFINE BUFFER tt-query-buffer FOR tt-query-buffer.

/****************************************************************************/

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

FIND tt-query-buffer
    WHERE   tt-query-buffer.qry-handle  = cur-query-handle AND
            tt-query-buffer.buffer-num  = cur-buffer-num
    NO-LOCK
    NO-ERROR.

RETURN(IF AVAILABLE tt-query-buffer
        THEN tt-query-buffer.BUFFER-HANDLE
        ELSE ?
        ).

END FUNCTION.
&ENDIF

/****************************************************************************/
FUNCTION get-qrymgr-buffer-field-set-name
/****************************************************************************
PURPOSE:

    Get the field set name for a given buffer-field-list specification.

PARAMETERS:

    cur-query-handle        Handle to the query this is associated with

    cur-buffer-field-list   List in buffer-postion.field-name comma-delimted
                            format (See is-qrymgr-break-condition() for specs)

    cur-flag-list           Optional flags to control behavior.

                            Flag                    Description
                            ---------------         ---------------------------------
                            "create-if-not-found"   Create a name if one can't be
                                                    found for the query and buffer
                                                    field list specified

*****************************************************************************/
     RETURNS CHARACTER

     (INPUT cur-query-handle        AS HANDLE,
      INPUT cur-buffer-field-list   AS CHARACTER,
      INPUT cur-flag-list           AS CHARACTER)           {&super-header}:

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Variables */

DEFINE VARIABLE is-create-if-not-found                  AS LOGICAL   NO-UNDO.

/****************************************************************************/
/* Local Buffers */

DEFINE BUFFER tt-named-buffer-field-set FOR tt-named-buffer-field-set.

/*****************************************************************************/

ASSIGN
    is-create-if-not-found = LOOKUP("create-if-not-found", cur-flag-list) > 0
    .

/****************************************************************************/

DO TRANSACTION: /* Transaction */

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

FIND tt-named-buffer-field-set
    WHERE tt-named-buffer-field-set.qry-handle          = cur-query-handle      AND
          tt-named-buffer-field-set.db-table-field-list = cur-buffer-field-list
    NO-LOCK
    NO-ERROR.

IF NOT AVAILABLE tt-named-buffer-field-set   AND
   is-create-if-not-found                   THEN

    DO:

    CREATE tt-named-buffer-field-set.

    ASSIGN
        cur-name-seq                                    = cur-name-seq      + 1
        tt-named-buffer-field-set.buffer-field-set-name = cur-local-name    + TRIM(STRING(cur-name-seq))

        tt-named-buffer-field-set.qry-handle            = cur-query-handle
        tt-named-buffer-field-set.db-table-field-list   = cur-buffer-field-list
        .

    FIND CURRENT tt-named-buffer-field-set
        NO-LOCK.

    END.

RETURN(IF AVAILABLE tt-named-buffer-field-set
        THEN tt-named-buffer-field-set.buffer-field-set-name
        ELSE "").

END.            /* Transaction */

END FUNCTION.
&ENDIF

/****************************************************************************/
FUNCTION is-qrymgr-break-condition
/****************************************************************************
PURPOSE:

    Determine if the desired break conditions have been met for a given
    query.

PARAMETERS:

    cur-query-handle            Handle to the query to check for break
                                conditions

    cur-db-table-field-list     List of db table fields to check values
                                for in the format buffer-num.field-name
                                ie

                                    1.field-one,2.field-two,3.field-three

                                NOTE: If no "." is specified, then this
                                parameter is assumed to be a buffer field
                                value set name.
                                
    cur-break-condition         Which break-by condition to check for.

                                Value       Description
                                ----------- --------------------------------
                                "FIRST-OF"  Is this the first record for the
                                            given list of field values

                                "LAST-OF"   Is this the last record for the
                                            given list of field values

*****************************************************************************/
     RETURNS LOGICAL

     (INPUT cur-query-handle        AS HANDLE,
      INPUT cur-db-table-field-list AS CHARACTER,
      INPUT cur-break-condition     AS CHARACTER)           {&super-header}:

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Buffers */

DEFINE BUFFER tt-query FOR tt-query.

/****************************************************************************/
/* Local Variables */

DEFINE VARIABLE is-break-not-met                AS LOGICAL          NO-UNDO.        /* ie, the data hasn't changed */
DEFINE VARIABLE cur-buffer-field-set-name       AS CHARACTER        NO-UNDO.

/*****************************************************************************/

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */


FIND tt-query
    WHERE tt-query.qry-handle = cur-query-handle
    NO-LOCK
    NO-ERROR.

/****************************************************************************/

ASSIGN
    is-break-not-met    = YES
    .

/****************************************************************************/
/* Find / create the named buffer field list for this specification */

IF NUM-ENTRIES(cur-db-table-field-list, ".") = 1 THEN               /* NO "." means this is a field value set name  */

    ASSIGN
        cur-buffer-field-set-name = cur-db-table-field-list
        .

ELSE                                                                /* Find the name and / or make one up           */

    ASSIGN
        cur-buffer-field-set-name = get-qrymgr-buffer-field-set-name(   cur-query-handle,
                                                                        cur-db-table-field-list,
                                                                        "create-if-not-found")
        .

/****************************************************************************/

DO WHILE VALID-HANDLE(cur-query-handle):            /* To provide a consistent exit point           */
                                                    /* and check the query's legit                  */

    IF cur-query-handle:QUERY-OFF-END THEN          /* If we're already off the end of a            */
        LEAVE.                                      /* query. Don't assume if we're first or last   */


        /* Make sure the field value set exists     */

    RUN create-qrymgr-named-buffer-field-set(               cur-query-handle,
                                                            cur-buffer-field-set-name,
                                                            cur-db-table-field-list,
                                                            "is-check").

        /* Save the current values for the fields   */
        /* in the current buffer-field-value-set    */

    RUN save-qrymgr-named-buffer-field-set-current-values(  cur-query-handle,
                                                            cur-buffer-field-set-name).

        /* Check for break conditions               */

    CASE cur-break-condition:
    WHEN "first-of" THEN                            /* If we're doing "First-of" then       */

        DO:

        cur-query-handle:GET-PREV(NO-LOCK).         /* Go back 1 result list entry          */

        IF cur-query-handle:QUERY-OFF-END THEN      /* Off the end?                         */
            DO:

            ASSIGN
                is-break-not-met = NO               /* Then we're first-of                  */
                .

            cur-query-handle:GET-FIRST(NO-LOCK).    /* restore position in result list      */

            END.

        ELSE

            DO:                                     /* Check current values against saved   */
                                                    /* values                               */

            RUN check-qrymgr-named-buffer-field-set-values( cur-query-handle,           /* Compare this list            */
                                                            cur-buffer-field-set-name,  /* With the current db values   */
                                                            OUTPUT is-break-not-met).   /* Did they match?              */

            cur-query-handle:GET-NEXT(NO-LOCK).     /* Reposition to starting point         */

            END.

        END.

    WHEN "last-of"  THEN

        DO:

        cur-query-handle:GET-NEXT(NO-LOCK).         /* Go forward a record                      */

        IF cur-query-handle:QUERY-OFF-END THEN      /* gone off the end of the result list?     */
            DO:

            ASSIGN
                is-break-not-met = NO               /* break conditions have been met           */
                .

            cur-query-handle:GET-LAST(NO-LOCK).     /* reposition back to where we were         */

            END.

        ELSE

            DO:                                     /* Check current buffer field value set     */
                                                    /* against the saved buffer field value set */

            RUN check-qrymgr-named-buffer-field-set-values( cur-query-handle,           /* Compare this list            */
                                                            cur-buffer-field-set-name,  /* With the current db values   */
                                                            OUTPUT is-break-not-met).   /* Did they match?              */

            cur-query-handle:GET-PREV(NO-LOCK).     /* restore position in the result list      */

            END.

        END.

    END CASE.

    LEAVE.

END.        /* DO WHILE VALID-HANDLE()  */

RETURN(NOT is-break-not-met).                       /* Negate break-not-met to find out if      */
END FUNCTION.                                       /* break was met or not                     */
&ENDIF

/****************************************************************************/
PROCEDURE create-qrymgr-named-buffer-field-set              {&super-header}:
/****************************************************************************
PURPOSE:

    Create a named field set for a given query and set of associated
    buffers and fields.

    NOTE:   If the current named field value set already exists,
            nothing is done

PARAMETERS:

    cur-query-handle            Current query handle

    cur-buffer-field-set-name   Name of this buffer field set

    cur-db-table-field-list     List of all table and fields to save
                                values from.

                                Specification is buffer-num.field-name format
                                ie:

                                    1.field-one,2.field-two,3.field-three


                                    
    cur-flag-list               Optional flags to control behavior.

                                Flag            Description
                                --------------- ---------------------------------
                                "is-check"      Don't update the list of fields in
                                                the field value list if an entry
                                                already exists for that named list

                                               
*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE    NO-UNDO.
DEFINE INPUT PARAMETER cur-buffer-field-set-name        AS CHARACTER NO-UNDO.
DEFINE INPUT PARAMETER cur-db-table-field-list          AS CHARACTER NO-UNDO.

DEFINE INPUT PARAMETER cur-flag-list                    AS CHARACTER NO-UNDO.

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Variables */

DEFINE VARIABLE cur-buffer-num                          AS INTEGER   NO-UNDO.
DEFINE VARIABLE cur-field-name                          AS CHARACTER NO-UNDO.
DEFINE VARIABLE cur-field-handle                        AS HANDLE    NO-UNDO.

DEFINE VARIABLE cur-cnt                                 AS INTEGER   NO-UNDO.

DEFINE VARIABLE cur-entry                               AS CHARACTER NO-UNDO.
DEFINE VARIABLE is-check                                AS LOGICAL   NO-UNDO.
DEFINE VARIABLE cur-num-entries                         AS INTEGER   NO-UNDO.

/****************************************************************************/
/* Local Buffers */

DEFINE BUFFER tt-query-buffer                   FOR tt-query-buffer.
DEFINE BUFFER tt-query-buffer-field-set-value   FOR tt-query-buffer-field-set-value.

/****************************************************************************/

ASSIGN
    is-check = LOOKUP("is-check", cur-flag-list) > 0
    .

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */
    
/****************************************************************************/
/* If this named field  set exists already, */
/* then get out                             */

IF is-check                                                                                         AND

    CAN-FIND(FIRST tt-query-buffer-field-set-value
            WHERE tt-query-buffer-field-set-value.qry-handle            = cur-query-handle          AND
                  tt-query-buffer-field-set-value.buffer-field-set-name = cur-buffer-field-set-name)    THEN
    LEAVE.

/****************************************************************************/

ASSIGN
    cur-num-entries = NUM-ENTRIES(cur-db-table-field-list)
    .

DO cur-cnt = 1 TO cur-num-entries:

        /* Get the current entry                            */

    ASSIGN
        cur-entry       = ENTRY(cur-cnt, cur-db-table-field-list)
        .

        /* Break it out into a buffer # and field name      */

    ASSIGN
        cur-buffer-num  = INTEGER(ENTRY(1, cur-db-table-field-list, "."))
        cur-field-name  =         ENTRY(2, cur-db-table-field-list, ".")
        .

    FIND tt-query-buffer
        WHERE tt-query-buffer.qry-handle = cur-query-handle AND
              tt-query-buffer.buffer-num = cur-buffer-num
        NO-LOCK
        NO-ERROR.

        /* If this buffer # doesn't exist for this query    */
        /* then skip this entry                             */

    IF NOT AVAILABLE tt-query-buffer THEN
        NEXT.

        /* Is this field in the list already?               */

    IF CAN-FIND(tt-query-buffer-field-set-value
            WHERE tt-query-buffer-field-set-value.qry-handle            = cur-query-handle          AND
                  tt-query-buffer-field-set-value.buffer-field-set-name = cur-buffer-field-set-name AND
                  tt-query-buffer-field-set-value.buffer-num            = cur-buffer-num)           THEN

            /* Then go to the next entry                    */

        NEXT.

        /* Get the field handle                             */

    ASSIGN
        cur-field-handle = tt-query-buffer.BUFFER-HANDLE:BUFFER-FIELD(cur-field-name)
        NO-ERROR
        .

        /* If there's no such field in the buffer           */
        /* then ignore this one                             */

    IF cur-field-handle = ? THEN
        NEXT.

        /* Create a new field-value-set entry               */

    CREATE tt-query-buffer-field-set-value.

    ASSIGN
        tt-query-buffer-field-set-value.qry-handle              = cur-query-handle
        tt-query-buffer-field-set-value.buffer-field-set-name   = cur-buffer-field-set-name

        tt-query-buffer-field-set-value.buffer-num              = cur-buffer-num
        tt-query-buffer-field-set-value.field-name              = cur-field-name
        tt-query-buffer-field-set-value.field-handle            = cur-field-handle
        .

END.

&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE save-qrymgr-named-buffer-field-set-current-values {&super-header}:
/****************************************************************************
PURPOSE:

    Save the value for a set of buffer fields to a set of temp-table
    records for a given query and name.

    If the named field-value set doesn't exist, then create it.

PARAMETERS:

    cur-query-handle            Current query handle

    cur-buffer-field-set-name   Name of this buffer field set

    cur-db-table-field-list     List of all table and fields to save
                                values from.

                                Specification is buffer-num.field-name format
                                ie:

                                    1.field-one,2.field-two,3.field-three


*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE    NO-UNDO.
DEFINE INPUT PARAMETER cur-buffer-field-set-name        AS CHARACTER NO-UNDO.

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Variables */

DEFINE VARIABLE cur-buffer-num                          AS INTEGER   NO-UNDO.
DEFINE VARIABLE cur-field-name                          AS CHARACTER NO-UNDO.
DEFINE VARIABLE cur-field-handle                        AS HANDLE    NO-UNDO.

DEFINE VARIABLE cur-cnt                                 AS INTEGER   NO-UNDO.

DEFINE VARIABLE cur-entry                               AS CHARACTER NO-UNDO.

/****************************************************************************/
/* Local Buffers */

DEFINE BUFFER tt-query-buffer                   FOR tt-query-buffer.
DEFINE BUFFER tt-query-buffer-field-set-value   FOR tt-query-buffer-field-set-value.

/****************************************************************************/

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

FOR EACH tt-query-buffer-field-set-value
    WHERE tt-query-buffer-field-set-value.qry-handle            = cur-query-handle AND
          tt-query-buffer-field-set-value.buffer-field-set-name = cur-buffer-field-set-name
    EXCLUSIVE-LOCK:

    ASSIGN
        tt-query-buffer-field-set-value.field-value = STRING(tt-query-buffer-field-set-value.field-handle:BUFFER-VALUE)
        .

END.

&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE check-qrymgr-named-buffer-field-set-values         {&super-header}:
/****************************************************************************
PURPOSE:

    Given a named buffer field set, see if the corresponding table buffer
    field values have changed from when they were last saved.

PARAMETERS:

    cur-query-handle            Query handle

    cur-buffer-field-set-name   Buffer field set name associated with this query

    is-field-values-matched     TRUE    if all the fields match the
                                        last-saved value

                                FALSE   if one of the fields doesn't match

*****************************************************************************/

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE    NO-UNDO.
DEFINE INPUT PARAMETER cur-buffer-field-set-name        AS CHARACTER NO-UNDO.

DEFINE OUTPUT PARAMETER is-field-values-matched         AS LOGICAL   NO-UNDO.

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Buffers */

DEFINE BUFFER tt-query-buffer-field-set-value   FOR tt-query-buffer-field-set-value.

/****************************************************************************/

ASSIGN
    is-field-values-matched = YES
    .

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

FOR EACH tt-query-buffer-field-set-value
    WHERE tt-query-buffer-field-set-value.qry-handle            = cur-query-handle          AND
          tt-query-buffer-field-set-value.buffer-field-set-name = cur-buffer-field-set-name
    NO-LOCK:

    ASSIGN
        is-field-values-matched = (tt-query-buffer-field-set-value.field-value = STRING(tt-query-buffer-field-set-value.field-handle:BUFFER-VALUE))
        .

    IF NOT is-field-values-matched THEN
        LEAVE.

END.

&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE delete-qrymgr-named-buffer-field-set              {&super-header}:
/****************************************************************************
PURPOSE:

    Delete a named buffer field set and all associated records from a query

PARAMETERS:

    cur-query-handle            Query handle

    cur-buffer-field-set-name   Name of the buffer field set

*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE    NO-UNDO.
DEFINE INPUT PARAMETER cur-buffer-field-set-name        AS CHARACTER NO-UNDO.

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Buffers */

DEFINE BUFFER tt-named-buffer-field-set         FOR tt-named-buffer-field-set.
DEFINE BUFFER tt-query-buffer-field-set-value   FOR tt-query-buffer-field-set-value.

/****************************************************************************/

DO TRANSACTION: /* Transaction */

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

FOR EACH tt-query-buffer-field-set-value
    WHERE tt-query-buffer-field-set-value.qry-handle            = cur-query-handle          AND
          tt-query-buffer-field-set-value.buffer-field-set-name = cur-buffer-field-set-name
    EXCLUSIVE-LOCK:

    DELETE tt-query-buffer-field-set-value.

END.

FIND tt-named-buffer-field-set
    WHERE tt-named-buffer-field-set.qry-handle              = cur-query-handle              AND
          tt-named-buffer-field-set.buffer-field-set-name   = cur-buffer-field-set-name
    EXCLUSIVE-LOCK
    NO-ERROR.

IF AVAILABLE tt-named-buffer-field-set THEN
    DELETE tt-named-buffer-field-set.

END.            /* Transaction */

&ENDIF
END PROCEDURE.

/****************************************************************************/
PROCEDURE dump-qrymgr-query                                 {&super-header}:
/****************************************************************************
PURPOSE:

    Dump information pertaining to a query

PARAMETERS:

    cur-query-handle    Query handle

*****************************************************************************/
/* Local Parameters */

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE    NO-UNDO.

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/

RUN get-qry-handle(INPUT-OUTPUT cur-query-handle).      /* Make sure this is a query handle */

FOR EACH tt-query-buffer
    WHERE tt-query-buffer.qry-handle = cur-query-handle

    NO-LOCK

    BY tt-query-buffer.qry-handle
    BY tt-query-buffer.buffer-num:

    MESSAGE "qry " tt-query-buffer.qry-handle:NAME                      SKIP
        tt-query-buffer.buffer-num tt-query-buffer.BUFFER-HANDLE:NAME   SKIP
        tt-query-buffer.join-spec                                       SKIP
        tt-query-buffer.search-spec                                     SKIP
        tt-query-buffer.join-search-spec
        VIEW-AS ALERT-BOX INFO BUTTON OK.


END.

&ENDIF
END PROCEDURE.

/****************************************************************************/
FUNCTION get-qrymgr-string-value
/****************************************************************************
PURPOSE:

    Given a searc spec and table, field, array spec,

    find the first or last value in a result list
        in the specified sort order
        and return it as a string

PARAMETERS:


    cur-table-name-handle-list  List of table / buffer handles to use in
                                getting this value (see create-qrymgr-query
                                for the specifics).

    cur-join-selection-spec     Selection criteria in open-qrymgr-query()
                                format

    cur-sort-spec               Sort criteria in open-qrymgr-query()
                                format

    cur-variable-field-spec     Name of the variable to return in
                                buffer-pos.variable-name[.array-element] format

                                1.field-name    ->  Buffer #1,field name
                                1.field-name.2  ->  Buffer #1,field name, element #2

    cur-record-not-found-value  Value to use if this search spec doesn't return
                                a value

    cur-flag-list               Flags to control behavior.

                                Flag            Description
                                --------------- ---------------------------------
                                "first"         Get the first value in the list     ["first" or "last" is mandatory]
                                "last"          Get the last value in the list      ["first" or "last" is mandatory]

*****************************************************************************/

     RETURNS CHARACTER

     (INPUT cur-table-name-handle-list  AS CHARACTER,

      INPUT cur-join-selection-spec     AS CHARACTER,
      INPUT cur-sort-spec               AS CHARACTER,

      INPUT cur-variable-field-spec     AS CHARACTER,

      INPUT cur-record-not-found-value  AS CHARACTER,

      INPUT cur-flag-list               AS CHARACTER)           {&super-header}:

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Variables */

DEFINE VARIABLE cur-query-handle                        AS HANDLE    NO-UNDO.

DEFINE VARIABLE cur-buffer-handle                       AS HANDLE    NO-UNDO.

DEFINE VARIABLE cur-buffer-num                          AS INTEGER   NO-UNDO.
DEFINE VARIABLE cur-buffer-variable-name                AS CHARACTER NO-UNDO.

DEFINE VARIABLE cur-buffer-field-handle                 AS HANDLE    NO-UNDO.

DEFINE VARIABLE cur-buffer-field-value                  AS CHARACTER NO-UNDO.
DEFINE VARIABLE cur-array-element                       AS INTEGER   NO-UNDO.

    /* Flags    */

DEFINE VARIABLE is-first-value                          AS LOGICAL   NO-UNDO.
DEFINE VARIABLE is-last-value                           AS LOGICAL   NO-UNDO.

/*****************************************************************************/
/* Setup things */

ASSIGN

        /* Flags for behavior                   */

    is-first-value              = LOOKUP("first",   cur-flag-list) > 0
    is-last-value               = LOOKUP("last",    cur-flag-list) > 0

        /* Buffer position and variable name    */

    cur-buffer-num              = INTEGER(ENTRY(1, cur-variable-field-spec, "."))
    cur-buffer-variable-name    = ENTRY(2, cur-variable-field-spec, ".")

        /* field array element                  */
        /* this defaults to element 0 if        */
        /* an element # isn't specified         */

    cur-array-element           = 0
    cur-array-element           = INTEGER(ENTRY(3, cur-variable-field-spec, ".")) WHEN NUM-ENTRIES(cur-variable-field-spec) > 2
    .

/****************************************************************************/

    /* Create the query         */

RUN create-qrymgr-query(cur-table-name-handle-list,
                        "",

                        OUTPUT cur-query-handle).

    /* Open it                  */

RUN open-qrymgr-query(  cur-query-handle,
                        cur-join-selection-spec,
                        cur-sort-spec,

                        IF is-last-value                /* If we want the last value                    */
                            THEN "invert-sort"          /* Then invert the sort order                   */
                                                        /* to avoid having to walk a large result list  */
                            ELSE "").

    /* Get us a record          */

cur-query-handle:GET-FIRST(NO-LOCK).

    /* Dereference the value    */

IF cur-query-handle:QUERY-OFF-END THEN

    ASSIGN
        cur-buffer-field-value      = cur-record-not-found-value
        .

ELSE

    ASSIGN
        cur-buffer-handle           = get-qrymgr-buffer-order-handle(cur-query-handle, cur-buffer-num)      /* Buffer handle    */

        cur-buffer-field-handle     = cur-buffer-handle:BUFFER-FIELD(cur-buffer-variable-name)              /* Field handle     */

        cur-buffer-field-value      = cur-buffer-field-handle:BUFFER-VALUE(cur-array-element)               /* Field value      */
        .

    /* Dump the query           */

RUN delete-qrymgr-query(cur-query-handle).

    /* Return the value         */

RETURN(cur-buffer-field-value).
END FUNCTION.
&ENDIF

/****************************************************************************/
FUNCTION get-qrymgr-decimal-value
/****************************************************************************
PURPOSE:

    Interface to get-qrymgr-string-value() that returns a decimal value

PARAMETERS:

    See get-qrymgr-string-value() for definitions

    NOTE: cur-record-not-found-value is DECIMAL data type

*****************************************************************************/
     RETURNS DECIMAL

     (INPUT cur-table-name-handle-list  AS CHARACTER,
                                                      
      INPUT cur-selection-spec          AS CHARACTER, 
      INPUT cur-sort-spec               AS CHARACTER, 
                                                      
      INPUT cur-variable-field-spec     AS CHARACTER, 
                                                      
      INPUT cur-record-not-found-value  AS DECIMAL,
                                                      
      INPUT cur-flag-list               AS CHARACTER)       {&super-header}:

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Variables */

DEFINE VARIABLE cur-decimal                             AS DECIMAL   NO-UNDO.

/*****************************************************************************/

RETURN(DECIMAL(get-qrymgr-string-value( cur-table-name-handle-list,
                                        cur-selection-spec,
                                        cur-sort-spec,
                                        cur-variable-field-spec,
                                        STRING(cur-record-not-found-value),
                                        cur-flag-list))).

END FUNCTION.
&ENDIF

/****************************************************************************/
FUNCTION get-qrymgr-integer-value
/****************************************************************************
PURPOSE:

    Interface to get-qrymgr-string-value() that returns a integer value

PARAMETERS:

    See get-qrymgr-string-value() for definitions

    NOTE: cur-record-not-found-value is INTEGER data type

*****************************************************************************/
     RETURNS INTEGER

     (INPUT cur-table-name-handle-list  AS CHARACTER,
                                                      
      INPUT cur-selection-spec          AS CHARACTER, 
      INPUT cur-sort-spec               AS CHARACTER, 
                                                      
      INPUT cur-variable-field-spec     AS CHARACTER, 
                                                      
      INPUT cur-record-not-found-value  AS INTEGER,
                                                      
      INPUT cur-flag-list               AS CHARACTER)       {&super-header}:

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Variables */

DEFINE VARIABLE cur-decimal                             AS DECIMAL   NO-UNDO.

/*****************************************************************************/

RETURN(INTEGER(get-qrymgr-string-value( cur-table-name-handle-list,
                                        cur-selection-spec,
                                        cur-sort-spec,
                                        cur-variable-field-spec,
                                        STRING(cur-record-not-found-value),
                                        cur-flag-list))).

END FUNCTION.
&ENDIF

/****************************************************************************/
FUNCTION get-qrymgr-date-value
/****************************************************************************
PURPOSE:

    Interface to get-qrymgr-string-value() that returns a date value

PARAMETERS:

    See get-qrymgr-string-value() for definitions

    NOTE: cur-record-not-found-value is DATE data type

*****************************************************************************/
     RETURNS DATE

     (INPUT cur-table-name-handle-list  AS CHARACTER,
                                                      
      INPUT cur-selection-spec          AS CHARACTER, 
      INPUT cur-sort-spec               AS CHARACTER, 
                                                      
      INPUT cur-variable-field-spec     AS CHARACTER, 
                                                      
      INPUT cur-record-not-found-value  AS DATE,
                                                      
      INPUT cur-flag-list               AS CHARACTER)       {&super-header}:

&IF DEFINED(super-header) = 0 &THEN
/****************************************************************************/
/* Local Variables */

DEFINE VARIABLE cur-decimal                             AS DECIMAL   NO-UNDO.

/*****************************************************************************/

RETURN(DATE(get-qrymgr-string-value(cur-table-name-handle-list,
                                    cur-selection-spec,
                                    cur-sort-spec,
                                    cur-variable-field-spec,
                                    STRING(INTEGER(cur-record-not-found-value)),    /* Turn the date into an integer        */
                                                                                    /* so if it comes back it'll be         */
                                                                                    /* accurately converted back to date    */
                                                                                    /* regardless of SESSION:DATE-FORMAT    */
                                    cur-flag-list))).

END FUNCTION.
&ENDIF

/****************************************************************************/
/*                                                                          */
/*                                                                          */
/*                          PRIVATE CODE                                    */
/*                                                                          */
/*                                                                          */
/****************************************************************************/
&IF DEFINED(super-header) = 0 &THEN

/****************************************************************************/
PROCEDURE set-tt-query-join-search-spec                             PRIVATE:
/****************************************************************************
PURPOSE:

    Set the join-search-spec for a tt-query-buffer record

PARAMETERS:

    BUFFER tt-query-buffer

*****************************************************************************/
/* Local Parameters */

DEFINE PARAMETER BUFFER tt-query-buffer FOR tt-query-buffer.

/****************************************************************************/

DO TRANSACTION: /* Transaction */

FIND CURRENT tt-query-buffer
    EXCLUSIVE-LOCK.

ASSIGN
    tt-query-buffer.join-search-spec =  tt-query-buffer.join-spec                   +

                                        (IF tt-query-buffer.join-spec   <> "" AND
                                            tt-query-buffer.search-spec <> ""
                                            THEN " AND "
                                            ELSE "")                                +

                                        tt-query-buffer.search-spec
    .


FIND CURRENT tt-query-buffer
    NO-LOCK.

END.            /* Transaction */

END PROCEDURE.

/****************************************************************************/
PROCEDURE add-query-condition                                       PRIVATE:
/****************************************************************************
PURPOSE:

    Conditionally add a condition to a query string

PARAMETERS:

    cur-query-string        INPUT/OUTPUT value for the current query string

    cur-buffer-num          Query buffer position to change

    cur-field-name          Name of the field in the buffer to use for the
                            operation
                        
    cur-query-boolean       Boolean operator.

                            NOTE: IF a "NOT" is placed in front of this,
                                  then the operation is changed from

                                    db.file = something

                                            to

                                    NOT(db.field = something)


                            NOTE:   If this is "APPEND", this assumes
                                    that cur-query-condition has the filter
                                    condition pre-constructed, so just
                                    append / AND it to the rest of the
                                    query string


    cur-query-condition     Value to check the buffer field name against

*****************************************************************************/
/* Local Parameters */

DEFINE INPUT-OUTPUT PARAMETER cur-query-string          AS CHARACTER NO-UNDO.

DEFINE INPUT PARAMETER cur-buffer-num                   AS INTEGER   NO-UNDO.
DEFINE INPUT PARAMETER cur-field-name                   AS CHARACTER NO-UNDO.

DEFINE INPUT PARAMETER cur-query-boolean                AS CHARACTER NO-UNDO.
DEFINE INPUT PARAMETER cur-query-condition              AS CHARACTER NO-UNDO.

/****************************************************************************/
/* Local Variables */

DEFINE VARIABLE cur-open-parens                         AS CHARACTER NO-UNDO.
DEFINE VARIABLE cur-close-parens                        AS CHARACTER NO-UNDO.

/****************************************************************************/
/* If this is a "NOT" then invert the condition */

IF cur-query-boolean BEGINS "NOT" THEN

    DO:

    ASSIGN
        cur-query-boolean   = TRIM(SUBSTRING(cur-query-boolean, 4))
        cur-open-parens     = "NOT ("
        cur-close-parens    = ")"
        .


    END.

ELSE

    ASSIGN
        cur-open-parens     = ""
        cur-close-parens    = ""
        .

/****************************************************************************/
/* add the condition                            */

ASSIGN
    cur-query-string =  cur-query-string                    +

                        (IF cur-query-string    <> "" AND
                            cur-query-condition <> ""
                            THEN " AND "
                            ELSE "")                        +

                        cur-open-parens                     +

                        (IF cur-query-boolean = "APPEND"                                                    /* The condition has the full spec          */
                            THEN    ""                                                                      /* So don't add the boolean / field name    */

                            ELSE    " &" + TRIM(STRING(cur-buffer-num)) + "." + cur-field-name  + " "  +    /* Construct the field / boolean            */
                                    cur-query-boolean                   + " "                               /* values                                   */
                            )                               +

                        (IF cur-query-condition = ?                                                         /* Handle missing values                    */
                            THEN "?"
                            ELSE cur-query-condition)       +

                        cur-close-parens
    .


END PROCEDURE.

/****************************************************************************/
PROCEDURE create-tt-query-buffer-record                             PRIVATE:
/****************************************************************************
PURPOSE:

    Create a standard buffer record associated with a given query

PARAMETERS:

    tt-query-buffer         BUFFER for the current tt-query-buffer table record
                           
    cur-query-handle        Handle to the query being managed
                           
    cur-buffer-num          Position of the buffer in the query

    cur-buffer-handle       Handle for the query buffer
                           
    is-created-local-flag   True if this buffer was created by the manager

*****************************************************************************/

/* Local Parameters */

DEFINE PARAMETER BUFFER tt-query-buffer                 FOR tt-query-buffer.

DEFINE INPUT PARAMETER cur-query-handle                 AS HANDLE    NO-UNDO.

DEFINE INPUT PARAMETER cur-buffer-num                   AS HANDLE    NO-UNDO.
DEFINE INPUT PARAMETER cur-buffer-handle                AS HANDLE    NO-UNDO.

DEFINE INPUT PARAMETER is-created-local-flag            AS LOGICAL   NO-UNDO.

/****************************************************************************/

CREATE tt-query-buffer.

ASSIGN
    tt-query-buffer.qry-handle          = cur-query-handle

    tt-query-buffer.buffer-num          = cur-buffer-num
    tt-query-buffer.buffer-handle       = cur-buffer-handle

    tt-query-buffer.is-created-local    = is-created-local-flag
    .

    /* Set some standardized values */

ASSIGN
    tt-query-buffer.db-name             = tt-query-buffer.buffer-handle:DBNAME
    tt-query-buffer.table-name          = tt-query-buffer.buffer-handle:NAME

    tt-query-buffer.db-table-name       = tt-query-buffer.db-name + "." +
                                          tt-query-buffer.table-name
    .

END PROCEDURE.


/****************************************************************************/
PROCEDURE control-browse                                            PRIVATE:
/****************************************************************************
PURPOSE:

    Given an operation and a tt-query buffer, control the browse
    associated with the query. This way updates to the query by changing
    the search condition, moving around, etc. won't cause flashing on
    the associated browse object

PARAMETERS:

    tt-query    BUFFER to the tt-query table

    cur-op      Operation to perform

                Value                   Description
                -----------             -------------------------------------------------

                "disable"               Disable the browse from refreshing

                "enable"                Enable the browse for refreshing

                "refresh"               Refresh the browse (ie, sync to the query)

                "Select-focused-row"    call select-focused-row() for the browse

                "Select-row-1"          call select-row(1) for the browse
    
*****************************************************************************/
/* Local Parameters */

DEFINE PARAMETER BUFFER tt-query FOR tt-query.

DEFINE INPUT PARAMETER cur-op                           AS CHARACTER NO-UNDO.

/****************************************************************************/
/* Local Variables */

DEFINE VARIABLE is-refreshable                          AS LOGICAL   NO-UNDO.

DEFINE VARIABLE cur-query-rowid                         AS ROWID     NO-UNDO
                                                            EXTENT 18.

DEFINE VARIABLE cur-buffer-num                          AS INTEGER   NO-UNDO.

/****************************************************************************/

IF NOT AVAILABLE tt-query                   OR              /* No query -                                           */
   NOT tt-query.is-browse-control           OR              /* Not in control of the browse                         */
   NOT VALID-HANDLE(tt-query.browse-handle) THEN            /* or the browse handle's not valid                     */

    LEAVE.                                                  /* Get out                                              */

/****************************************************************************/

DO TRANSACTION: /* Transaction */

FIND CURRENT tt-query
    EXCLUSIVE-LOCK.

/****************************************************************************/
    
CASE cur-op:

/****************************************************************************/
WHEN "deselect-rows"            THEN

    DO:

    tt-query.browse-handle:DESELECT-ROWS() NO-ERROR.        /* Deselect all rows                                    */

    END.

/****************************************************************************/
WHEN "disable"                  THEN

    DO:

    ASSIGN                                                  /* set it to not refresh on query change                */
        tt-query.browse-handle:REFRESHABLE = NO             /* this deals with the flashing browse problem          */
        .                                                   /* Unfortunately this causes other problems             */

    END.

/****************************************************************************/
WHEN "enable"                   THEN

    DO:

    RUN activate-browse(BUFFER tt-query).

    END.
    
/****************************************************************************/
WHEN "refresh"                  THEN

    DO:

    ASSIGN
        is-refreshable  = tt-query.browse-handle:REFRESHABLE
        .

    RUN activate-browse(BUFFER tt-query).

        /* Put the "refreshable" setting back   */

    IF is-refreshable <> tt-query.browse-handle:REFRESHABLE THEN
        ASSIGN
            tt-query.browse-handle:REFRESHABLE  = is-refreshable
            .

    END.

/****************************************************************************/
WHEN "select-row-1"             THEN

    DO:

    IF tt-query.browse-handle:REFRESHABLE THEN

        tt-query.browse-handle:SELECT-ROW(1)        NO-ERROR.

    ELSE

        ASSIGN
            tt-query.is-select-focused-row-pending  = NO
            tt-query.is-select-row-1-pending        = YES
            .

    END.

/****************************************************************************/
WHEN "select-focused-row"       THEN

    DO:

    IF tt-query.browse-handle:REFRESHABLE THEN

        tt-query.browse-handle:SELECT-FOCUSED-ROW() NO-ERROR.

    ELSE

        ASSIGN
            tt-query.is-select-focused-row-pending  = YES
            tt-query.is-select-row-1-pending        = NO
            .

    END.

END CASE.

FIND CURRENT tt-query
    NO-LOCK.

END.            /* Transaction */

END PROCEDURE.

/****************************************************************************/
PROCEDURE activate-browse                                           PRIVATE:
/****************************************************************************
PURPOSE:

    We're here if the user's going from a "disabled" (non-refreshable)
    browse to an enabled ("refreshable") browse.

    Make sure whatever needs to be done when that happens gets done.


    Part of the reason for this is that if you're setting REFRESHABLE
    from NO to YES for a browse, it doesn't sync to the query but completely
    messes it up.

    The solution to this is
        1) save the current ROWIDs for the query,
        2) set REFRESHABLE to YES,
        3) then put the query back where it originally was.

    this results in the the browse being left positioned *exactly* where
    it belongs, rather than some random place as it was before.

PARAMETERS:

    tt-query        BUFFER for a tt-query record.
                    This buffer is assumed always present

*****************************************************************************/
/* Local Parameters */

DEFINE PARAMETER BUFFER tt-query        FOR tt-query.

/****************************************************************************/
/* Local Variables */

DEFINE VARIABLE cur-rowid-array                     AS ROWID    NO-UNDO EXTENT 18.
DEFINE VARIABLE cur-buffer-handle                   AS HANDLE   NO-UNDO.
DEFINE VARIABLE cur-first-buffer-handle             AS HANDLE   NO-UNDO.

DEFINE VARIABLE cur-int                             AS INTEGER  NO-UNDO.
DEFINE VARIABLE is-logical                          AS LOGICAL  NO-UNDO.

DEFINE VARIABLE is-refresh-required                 AS LOGICAL  NO-UNDO.
DEFINE VARIABLE is-focused-row-selected             AS LOGICAL  NO-UNDO.
DEFINE VARIABLE is-row-1-selected                   AS LOGICAL  NO-UNDO.

/****************************************************************************/
/* Local Buffers */

DEFINE BUFFER tt-query-lu               FOR tt-query.

/****************************************************************************/

ASSIGN
    is-refresh-required = YES               /* Make this a safe assumption  */
    is-row-1-selected   = NO                /* Make this a safe assumption  */
    .

/****************************************************************************/
/* If we're going to REFRESHABLE = YES, then        */
/* do some finagling to make it work as it should   */

IF NOT tt-query.browse-handle:REFRESHABLE THEN
    DO:

    ASSIGN
        cur-rowid-array = ?
        .

        /* Save the query position                          */

    DO cur-int = 1 TO tt-query.qry-handle:NUM-BUFFERS:

        ASSIGN
            cur-buffer-handle           = tt-query.qry-handle:GET-BUFFER-HANDLE(cur-int)
            cur-first-buffer-handle     = cur-buffer-handle             WHEN cur-int = 1
            .

        ASSIGN
            cur-rowid-array[cur-int]    = cur-buffer-handle:ROWID       WHEN cur-buffer-handle:AVAILABLE
            .

    END.

        /* Refresh the browse from the query                */

    ASSIGN
        tt-query.browse-handle:REFRESHABLE = YES
        .

        /* IF we have one, then go back to the original     */
        /* result-list entry if the browse messed it up     */

    IF cur-rowid-array[1] <> ?                              AND
       cur-rowid-array[1] <> cur-first-buffer-handle:ROWID  THEN

        tt-query.qry-handle:REPOSITION-TO-ROWID(cur-rowid-array)
            NO-ERROR.

        /* Couldn't go to that result list entry            */
        /* or the query wasn't open - try to go to row 1    */

    IF ERROR-STATUS:ERROR       OR
       cur-rowid-array[1] = ?   THEN

        tt-query.qry-handle:GET-FIRST(NO-LOCK).

    ASSIGN
        is-refresh-required = NO        /* Because the browse just got redone    */
        .
        
    END.

/****************************************************************************/
/* Which row of the viewport            */
/* iteration to go to?                  */

ASSIGN
    is-row-1-selected = tt-query.browse-handle:IS-ROW-SELECTED(1)
    NO-ERROR
    .

ASSIGN
    is-row-1-selected   = NO                    WHEN ERROR-STATUS:ERROR
    .

IF tt-query.is-select-row-1-pending             AND
   NOT is-row-1-selected                        THEN

    tt-query.browse-handle:SELECT-ROW(1)        NO-ERROR.

/****************************************************************************/
/* Set the focused row flag?            */

ASSIGN
    is-focused-row-selected = tt-query.browse-handle:FOCUSED-ROW-SELECTED
    NO-ERROR
    .

ASSIGN
    is-focused-row-selected = NO                WHEN ERROR-STATUS:ERROR
    .

IF tt-query.is-select-focused-row-pending       AND
   NOT is-focused-row-selected                  THEN

    tt-query.browse-handle:SELECT-FOCUSED-ROW() NO-ERROR.

/****************************************************************************/
/* Refresh the browse on the screen     */

IF is-refresh-required THEN
    tt-query.browse-handle:REFRESH()            NO-ERROR.

/****************************************************************************/
/* Clear these flags                    */

ASSIGN
    tt-query.is-select-focused-row-pending  = NO
    tt-query.is-select-row-1-pending        = NO
    .

END PROCEDURE.

/****************************************************************************/
PROCEDURE get-qry-handle                                            PRIVATE:
/****************************************************************************
PURPOSE:

    Given a browse / query handle, de-reference this to a query handle

PARAMETERS:

    cur-handle      INPUT-OUTPUT handle for query / browse handle

*****************************************************************************/
/* Local Parameters */

DEFINE INPUT-OUTPUT PARAMETER cur-handle                AS HANDLE    NO-UNDO.

/****************************************************************************/

IF NOT VALID-HANDLE(cur-handle) THEN
    LEAVE.
    
CASE cur-handle:TYPE:
WHEN "browse"   THEN

    ASSIGN
        cur-handle = cur-handle:QUERY
        .

END CASE.

END PROCEDURE.

/****************************************************************************/
PROCEDURE construct                                                 PRIVATE:
/****************************************************************************
PURPOSE:


PARAMETERS:
****************************************************************************/

    /* Initializations          */

ASSIGN
    cur-local-name  =   THIS-PROCEDURE:FILE-NAME + "-" +    /* pseudo-random name for internal use  */
                        TRIM(STRING(THIS-PROCEDURE))

    cur-name-seq    =   0                                   /* Sequence number for where */
    .

END PROCEDURE.


/****************************************************************************/
PROCEDURE destruct                                                  PRIVATE:
/****************************************************************************
PURPOSE:

   Cleanup when shutting this SP down - make sure all the stuff this
   procedure's made goes away when this procedure goes away.

   "Just in case" we clear out the widget-pool of all created objects
   as well. The widget-pool should be empty when the DELETE statement
   is run.

PARAMETERS:

    None. This goes through the entire tt-query table and deletes all
    the outstanding queries.

*****************************************************************************/

FOR EACH tt-query:

    RUN delete-qrymgr-query(tt-query.qry-handle).

END.

DELETE WIDGET-POOL.

END PROCEDURE.

/****************************************************************************/
&ENDIF                                                  /* PRIVATE CODE     */

